# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

import api
from api.core import API, secure
from api.security import checkPermissions

from .. import defaultListHandler, defaultObjectHandler

from cli import Cli

from services import Service

from tools.config import Config
from tools.license import getLicense, updateCertificate
from tools.permissions import SystemAdminPermission, SystemAdminROPermission
from tools.dnsHealth import getHostByName
from tools.misc import callUpdateScript

import json
import os
import psutil
import requests
import shlex
import time

from datetime import datetime
from flask import jsonify, make_response, request
from io import StringIO


@API.route(api.BaseRoute+"/system/dashboard", methods=["GET"])
@secure()
def getDashboard():
    checkPermissions(SystemAdminROPermission())
    disks = []
    for disk in psutil.disk_partitions():
        try:
            usage = psutil.disk_usage(disk.mountpoint)
            stat = {"percent": usage.percent, "total": usage.total, "used": usage.used, "free": usage.free}
            stat["device"] = disk.device
            stat["mountpoint"] = disk.mountpoint
            stat["filesystem"] = disk.fstype
            disks.append(stat)
        except:
            pass
    cpu = psutil.cpu_times_percent()
    cpuPercent = dict(user=cpu.user, system=cpu.system, io=cpu.iowait, interrupt=cpu.irq+cpu.softirq, steal=cpu.steal,
                      idle=cpu.idle)
    vm = psutil.virtual_memory()
    memory = dict(percent=vm.percent, total=vm.total, used=vm.used, buffer=vm.buffers, cache=vm.cached, free=vm.free,
                  available=vm.available)
    sm = psutil.swap_memory()
    swap = dict(percent=sm.percent, total=sm.total, used=sm.used, free=sm.free)
    return jsonify(disks=disks,
                   load=os.getloadavg(),
                   cpuPercent=cpuPercent,
                   memory=memory,
                   swap=swap,
                   booted=datetime.fromtimestamp(psutil.boot_time()).strftime("%Y-%m-%d %H:%M:%S"))


@API.route(api.BaseRoute+"/system/dashboard/services", methods=["GET"])
@secure()
def getDashboardServices():
    checkPermissions(SystemAdminROPermission())
    known = Config["options"]["dashboard"]["services"]
    if len(known) == 0:
        return jsonify(services=[])
    with Service("systemd") as sysd:
        units = sysd.getServices(*(service["unit"] for service in known))
        for service in known:
            if service["unit"] not in units:
                continue
            units[service["unit"]]["name"] = service.get("name", service["unit"].replace(".service", ""))
        return jsonify(services=list(units.values()))


@API.route(api.BaseRoute+"/system/dashboard/services/<unit>", methods=["GET"])
@secure()
def getDashboardService(unit):
    checkPermissions(SystemAdminROPermission())
    for service in Config["options"]["dashboard"]["services"]:
        if service["unit"] == unit:
            break
    else:
        return jsonify(message="Unknown unit '{}'".format(unit)), 400
    with Service("systemd") as sysd:
        unit = sysd.getServices(service["unit"])[service["unit"]]
        unit["name"] = service.get("name", service["unit"].replace(".service", ""))
        unit["unit"] = service["unit"]
        return jsonify(unit)


@API.route(api.BaseRoute+"/system/dashboard/services/<unit>/<action>", methods=["POST"])
@secure()
def signalDashboardService(unit, action):
    checkPermissions(SystemAdminPermission())
    if action not in ("start", "stop", "restart", "reload", "enable", "disable"):
        return jsonify(message="Invalid action"), 400
    if unit not in (service["unit"] for service in Config["options"]["dashboard"]["services"]):
        return jsonify(message="Unknown unit '{}'".format(unit)), 400
    with Service("systemd") as sysd:
        _, msg = sysd.run(action, unit)
        return jsonify(message=msg or "Success"), 500 if msg else 201


def dumpLicense():
    License = getLicense()
    try:
        from orm.users import Users
        currentUsers = Users.count()
    except:
        currentUsers = None
    return jsonify(product=License.product,
                   maxUsers=License.users,
                   notBefore=License.notBefore.strftime("%Y-%m-%d %H:%M:%S"),
                   notAfter=License.notAfter.strftime("%Y-%m-%d %H:%M:%S"),
                   currentUsers=currentUsers,
                   certificate="/api/v1/system/license/certificate.pem" if License.cert is not None else None)


@API.route(api.BaseRoute+"/system/license", methods=["GET"])
@secure()
def getLicenseInfo():
    checkPermissions(SystemAdminROPermission())
    return dumpLicense()


@API.route(api.BaseRoute+"/system/license/certificate.pem", methods=["GET"])
@secure()
def getLicenseFile():
    checkPermissions(SystemAdminPermission())
    License = getLicense()
    if License.file is None:
        return jsonify(message="No license installed"), 404
    response = make_response(License.file)
    response.headers.set("Content-Type", "application/x-pem-file")
    return response


@API.route(api.BaseRoute+"/system/license", methods=["PUT"])
@secure()
def updateLicense():
    checkPermissions(SystemAdminPermission())
    error = updateCertificate(request.get_data())
    if error:
        return jsonify(message=error), 400
    return dumpLicense()


@API.route(api.BaseRoute+"/system/license/creds", methods=["GET"])
@secure()
def getLicenseSubscription():
    checkPermissions(SystemAdminPermission())
    try:
        with open(Config["options"]["subscriptionFile"]) as file:
            creds = file.readline()
        username, password = creds.split(":")
        return jsonify(username=username, password=password)
    except KeyError:
        return jsonify(message="Subscription detail storage path not configured"), 503
    except Exception:
        return jsonify(username=None, password=None)


@API.route(api.BaseRoute+"/system/license/creds", methods=["PUT"])
@secure()
def setLicenseSubscription():
    checkPermissions(SystemAdminPermission())
    data = request.get_json(silent=True)
    if data is None or "username" not in data or "password" not in data:
        return jsonify(message="Incomplete data"), 400
    try:
        with open(Config["options"]["subscriptionFile"], "w") as file:
            file.write(":".join((data["username"], data["password"])))
        return jsonify(message="okidoki")
    except KeyError:
        return jsonify(message="Subscription detail storage path not configured"), 503
    except PermissionError as err:
        return jsonify(message="Could not save credentials: "+err.args[1]), 500
    except FileNotFoundError:
        return jsonify(message="Could not save credentials: no such file or directory"), 500
    except Exception as err:
        API.logger.error("Could not save credentials:"+type(err).__name__+": "+" - ".join(str(arg) for arg in err.args))
        return jsonify(message="Could not save credentials"), 500


@API.route(api.BaseRoute+"/system/antispam/<path:path>", methods=["GET"])
@secure()
def rspamdProxy(path):
    checkPermissions(SystemAdminROPermission())
    conf = Config["options"]
    if path not in conf.get("antispamEndpoints", ("stat", "graph", "errors")):
        return jsonify(message="Endpoint not allowed"), 403
    try:
        res = requests.get(conf["antispamUrl"]+"/"+path, request.args, stream=True)
    except BaseException as err:
        API.logger.error(type(err).__name__+": "+" - ".join(str(arg) for arg in err.args))
        return jsonify(message="Failed to connect to antispam"), 503
    return res.raw.read(), res.status_code, res.headers.items()


@API.route(api.BaseRoute+"/system/vhostStatus", methods=["GET"])
@secure()
def vhostStatusList():
    checkPermissions(SystemAdminROPermission())
    return jsonify(data=list(Config["options"].get("vhosts", {}).keys()))


@API.route(api.BaseRoute+"/system/vhostStatus/<path:host>", methods=["GET"])
@secure()
def vhostStatus(host):
    checkPermissions(SystemAdminROPermission())
    conf = Config["options"].get("vhosts", {})
    if host not in conf:
        return jsonify(message="VHost not found"), 404
    try:
        res = requests.get(conf[host], stream=True)
    except BaseException as err:
        API.logger.error(type(err).__name__+": "+" - ".join(str(arg) for arg in err.args))
        return jsonify(message="Failed to connect to vhost"), 503
    return res.raw.read(), res.status_code, res.headers.items()


@API.route(api.BaseRoute+"/system/cli", methods=["POST"])
@secure()
def cliOverRest():
    checkPermissions(SystemAdminPermission())
    if Config["options"].get("disableRemoteCli"):
        return jsonify(message="Remote CLI disabled by config"), 400
    params = request.get_json(silent=True)
    if params is None:
        return jsonify(message="Missing parameters"), 400
    if "command" not in params:
        return jsonify(message="Missing command"), 400
    mode = params.get("mode", "exec")
    stdout = StringIO()
    fs = params.get("fs")
    cli = Cli(mode="adhoc", stdin=None, stdout=stdout, color=params.get("color", False), fs=fs)
    if mode == "complete":
        return jsonify(completions=cli.complete(params["command"]))
    API.logger.info("Executing CLI command '{}'".format(params["command"]))
    result = 0
    try:
        result = cli.execute(shlex.split(params["command"]), secure=False)
    except SystemExit:
        pass
    except Exception as err:
        return jsonify(message="{} ({})".format(type(err).__name__, " - ".join(str(arg) for arg in err.args))), 500
    cli.closeFiles()
    return jsonify(code=result, stdout=stdout.getvalue(), fs=cli.fs)


@API.route(api.BaseRoute+"/system/sync/top", methods=["GET"])
@secure()
def syncTop():
    checkPermissions(SystemAdminROPermission())
    sync = Config["sync"]
    expUpd = sync.get("topExpireUpdate", 120)
    expEnd = sync.get("topExpireEnded", 20)
    fupd = int(request.args.get("filterUpdated", expUpd))
    fend = int(request.args.get("filterEnded", expEnd))
    now = int(time.mktime(time.localtime()))
    with Service("redis") as r:
        r.set(sync.get("topTimestampKey", "grommunio-sync:topenabledat"), now)
        hdata = r.hgetall(sync.get("topdataKey", "grommunio-sync:topdata"))
        if hdata is None:
            return jsonify(data=[], maxUpdated=expUpd, maxEnded=expEnd)
        data = []
        remove = []
        for key, value in hdata.items():
            try:
                value = json.loads(value)
                if (value["ended"] != 0 and now-value["ended"] > expEnd) or \
                   now-value["update"] > expUpd:
                    remove.append(key)
                elif not (value["ended"] != 0 and now-value["ended"] > fend or now-value["update"] > fupd):
                    data.append(value)
            except Exception as err:
                API.logger.warning(type(err).__name__+": "+str(err.args))
        if len(remove) > 0:
            r.hdel(sync.get("topdataKey", "grommunio-sync:topdata"), *remove)
    return jsonify(data=data)


@API.route(api.BaseRoute+"/system/servers", methods=["GET", "POST"])
@secure(requireDB=105)
def getServersList():
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.misc import Servers
    return defaultListHandler(Servers)


@API.route(api.BaseRoute+"/system/servers/<int:ID>", methods=["GET", "PATCH", "DELETE"])
@secure(requireDB=105)
def getServerObject(ID):
    checkPermissions(SystemAdminROPermission() if request.method == "GET" else SystemAdminPermission())
    from orm.misc import Servers
    if request.method == "DELETE":
        server = Servers.query.filter(Servers.ID == ID).first()
        if server is not None and server.users+server.domains != 0:
            return jsonify(message="Cannot delete server with users or domains"), 400
    return defaultObjectHandler(Servers, ID, "")


@API.route(api.BaseRoute+"/system/servers/dnsCheck", methods=["GET"])
@secure(requireDB=105)
def getServersDNSChecks():
    checkPermissions(SystemAdminROPermission())
    from orm.misc import Servers
    objects = Servers.query.all()
    uniqueHostnames = [o.hostname for o in objects]
    uniqueExternalNames = set(o.extname for o in objects)

    hostsDNSCheck = {hostname: getHostByName(hostname) for hostname in uniqueHostnames}
    extDNSCheck = {extname: getHostByName(extname) for extname in uniqueExternalNames}

    return jsonify({"host": hostsDNSCheck, "ext": extDNSCheck})


@API.route(api.BaseRoute+"/system/updates/<string:command>", methods=["POST"])
@secure()
def checkForUpdates(command):
    repo = request.args["repo"] if "repo" in request.args else "supported"
    checkPermissions(SystemAdminPermission())
    pid = callUpdateScript(command or "check", repo)
    return jsonify({"pid": pid})



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

class Permissions:
    """Central Permissions class.

    Functions as a permission factory and provides easy permission checking.
    """

    preg = {}
    pregr = {}

    def __init__(self, *args):
        """Initialize permission object.

        Parameters
        ----------
        *args : Permission
            Permissions held by the object
        """
        self.permissions = args

    @classmethod
    def fromDB(cls, permissionsData=[]):
        """Initialize from database objects.

        Parameters
        ----------
        permissionsData : Iterable, optional
            List of objects with `name` and `param` attributes. The default is [].

        Returns
        -------
        Permissions
            Permissions object with permissions loaded from database
        """
        return Permissions(*(cls.preg[pData.permission](pData.params)
                             for pData in permissionsData if pData.permission in cls.preg))

    @classmethod
    def dump(cls, permission):
        """Create a permission representation that can be loaded with `load`.

        Parameters
        ----------
        permission : PermissionBase
            Permission to serialize

        Returns
        -------
        str
            String representation or None if serializaiton fails
        """
        if permission is None:
            return None
        try:
            import json
            data = {"permission": cls.pregr[permission] if type(permission) == type else cls.pregr[type(permission)]}
            if type(permission) != type and permission.params is not None:
                data["params"] = permission.params
            return json.dumps(data, separators=(",", ":"))
        except Exception:
            return None

    @classmethod
    def load(cls, permission=None, default=None):
        """Load a permission object.

        If permission is None, the return value is None, regardless of the default.

        Parameters
        ----------
        permission: str, optional
            Permission data, as created by dump(). The default is None.
        default : PermissionBase, optional
            Default permission if loading fails. The default is None.

        Returns
        -------
        PermissionBase
            A single permission constructed from the serialization
        """
        if permission is None:
            return
        try:
            import json
            data = json.loads(permission)
            return cls.preg[data["permission"]](data.get("params"))
        except Exception:
            return None

    def has(self, permission):
        """
        Check if permission is represented by permissions object.

        Parameters
        ----------
        permission : Permission
            Requested permission

        Returns
        -------
        bool
            True if the requested permission is represented, False otherwise
        """
        return permission is None or any(perm.permits(permission) for perm in self.permissions)

    def __contains__(self, permission):
        """Convenience alias for `has`.

        Parameters
        ----------
        permission : Permission
            Requested permission

        Returns
        -------
        bool
            True if the requested permission is represented, False otherwise
        """
        return self.has(permission)

    def __iter__(self):
        """Return permission iterator

        Returns
        -------
        iterator
            Iterator that iterates over the contained permission objects.
        """
        return self.permissions.__iter__()

    @classmethod
    def register(cls, name):
        """Class decorator to register a permission at the factory.

        Parameters
        ----------
        name : str
            Name of the permission

        Returns
        -------
        function
            Decorator function
        """
        def inner(obj):
            cls.preg[name] = obj
            cls.pregr[obj] = name
            return obj
        return inner

    @classmethod
    def knownPermissions(cls):
        """List names of registered permissions.

        Returns
        -------
        tuple
            List of registered permissions
        """
        return tuple(cls.preg.keys())

    @classmethod
    def create(cls, name, params=None):
        """Create permission by name.

        Parameters
        ----------
        name : str
            Name of the permission
        params : Any, optional
            Parameters passed to the permission. The default is None.

        Raises
        ------
        KeyError
            No permission with this name exists

        Returns
        -------
        Permission
            New permission.
        """
        if name not in cls.preg:
            raise KeyError("Unknown permission '{}'".format(name))
        return cls.preg[name](params)

    @staticmethod
    def sysadmin():
        """Create Permissions object with system admin permissions.

        Returns
        -------
        Permissions
            Permissions object with system admin permissions.
        """
        return Permissions(SystemAdminPermission())

    def capabilities(self):
        """Return set of capabilities from all represented permissions.

        Returns
        -------
        set
            Union of capabilities from all permissions
        """
        return set.union(*(permission.capabilities() for permission in self.permissions))


class PermissionBase:
    """Base class for permissions.

    Implements `permits` method, automatically dispatching the correct, permission specific, `_permits` method.
    """

    def __init__(self, *args, **kwargs):
        """Default constructor."""
        pass

    def permits(self, permission):
        """Check if `permission` is represented by the object.

        Checks if self is an instance of `permission` and, if so, dispatches the `_permits` method of the correct base class.

        Parameters
        ----------
        permission : PermissionBase
            Permission to check for.

        Returns
        -------
        bool
            True if `permission` is represented by this object, False otherwise
        """
        permission_t = type(permission)
        if type(self) == permission_t:
            self._permits(permission)
        if isinstance(self, permission_t):
            return permission_t._permits(self, permission)
        return False

    def _permits(self, permission):
        """Check if self contains requeusted permission.

        Called by the default `permits` method if `permission` and self are of the same type.
        By default, having the permission is sufficient, but it may be overridden to check permission parameters.

        Parameters
        ----------
        permission : same as type(self)
            Requested permission

        Returns
        -------
        bool:
            Whether permission matches
        """
        return True

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Empty set
        """
        return set()

    def __repr__(self):
        """String representation.

        Generic implementation providing class name with appended braces.

        Returns
        -------
        str
            String representation
        """
        return type(self).__name__+"()"

    @property
    def params(self):
        """Parameters relevant for the permission."""
        return None


@Permissions.register("SystemAdmin")
class SystemAdminPermission:
    """System admin permission.

    Permits every action by default.
    """

    __capcache = None

    def __init__(self, *args, **kwargs):
        """Initialize permission.

        Parameters
        ----------
        *args : Any
            Ignored.
        **kwargs : Any
            Ignored.
        """
        pass

    def permits(self, permission):
        """Return True.

        Parameters
        ----------
        permission : PermissionBase
            Permission to check

        Returns
        -------
        bool
            Always True
        """
        return True

    def __repr__(self):
        """Return string representation."""
        return "SystemAdminPermission()"

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Set containing all known capabilities.
        """
        if cls.__capcache is None:
            cls.__capcache = {"SystemAdminWrite"}
            for permission in Permissions.preg.values():
                cls.__capcache |= permission.capabilities()
        return cls.__capcache

    @property
    def params(self):
        """Parameters relevant for the permission."""
        return None


@Permissions.register("DomainAdminRO")
class DomainAdminROPermission(PermissionBase):
    """Permission class representing read only permissions for a domain.

    Can represent permission for a specific domain, or domains in general (when parameter is '*').

    Note that the special parameter '*' is permissive in both directions:
    Requesting a DomainAdminPermission with parameter '*' and
    requesting DomainAdminPermission for a specific domain from a permission with parameter '*' will both return True.
    """

    def __init__(self, domainID):
        """Initialize domain admin permission.

        Parameters
        ----------
        domainID : int or '*'
            Domain ID this permission is for, or '*' to match all domains

        Raises
        ------
        ValueError
            `domainID` is neither an integer nor special identifier '*'
        """
        if domainID != "*" and not isinstance(domainID, int):
            raise ValueError("DomainAdminROPermission parameter must be integer or '*'")
        self.__domain = domainID

    def _permits(self, permission):
        """Check if permission is represented.

        Parameters
        ----------
        permission : DomainAdminPermission
            Permission to check for equality

        Returns
        -------
        bool
            True if domain IDs match or either is a wildcard domain, False otherwise
        """
        return "*" in (self.__domain, permission.__domain) or self.__domain == permission.__domain

    def __repr__(self):
        """Return string representation."""
        return "DomainAdminROPermission({})".format(repr(self.__domain))

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Set containing "DomainAdmin" capability.
        """
        return {"DomainAdminRead"} | super().capabilities()

    @property
    def domainID(self):
        """Return domain parameter."""
        return self.__domain

    @property
    def params(self):
        """Parameters relevant for the permission."""
        return self.__domain


@Permissions.register("DomainAdmin")
class DomainAdminPermission(DomainAdminROPermission):
    """Permission class representing admin permissions for a domain.

    Can represent permission for a specific domain, or domains in general (when parameter is '*').

    Note that the special parameter '*' is permissive in both directions:
    Requesting a DomainAdminPermission with parameter '*' and
    requesting DomainAdminPermission for a specific domain from a permission with parameter '*' will both return True.
    """

    def __init__(self, domainID):
        """Initialize domain admin permission.

        Parameters
        ----------
        domainID : int or '*'
            Domain ID this permission is for, or '*' to match all domains

        Raises
        ------
        ValueError
            `domainID` is neither an integer nor special identifier '*'
        """
        if domainID != "*" and not isinstance(domainID, int):
            raise ValueError("DomainAdminPermission parameter must be integer or '*'")
        DomainAdminROPermission.__init__(self, domainID)

    def __repr__(self):
        """Return string representation."""
        return "DomainAdminPermission({})".format(repr(self.domainID))

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Set containing "DomainAdmin" capability.
        """
        return {"DomainAdminWrite"} | super().capabilities()


@Permissions.register("SystemAdminRO")
class SystemAdminROPermission(DomainAdminROPermission):
    """Permission class representing read-only system admin permissions.

    The read-only system admin has access to all the data a normal system admin has,
    but cannot modify anything.
    """
    def __init__(self, *args, **kwargs):
        """Initialize permission.

        Parameters
        ----------
        *args : Any
            Ignored.
        **kwargs : Any
            Ignored.
        """
        DomainAdminROPermission.__init__(self, "*")

    def __repr__(self):
        """Return string representation."""
        return "SystemAdminROPermission()"

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by the permission.

        Returns
        -------
        set
            Set containing "SystemAdminRO" and "DomainAdminRO" capabilities.
        """
        return {"SystemAdminRead"} | super().capabilities()

    @property
    def params(self):
        """Parameters relevant for the permission."""
        return None


@Permissions.register("OrgAdmin")
class OrgAdminPermission(PermissionBase):
    """Permission class representing admin permissions for an organization.

    An organization admin automatically has DomainAdminPermission for each domain belonging to an organization.
    Additionally, an organization admin can modify and delete domains.

    Can represent permission for a specific organization, or organizations in general (when parameter is '*').

    Note that the special parameter '*' is permissive in both directions:
    Requesting a DomainAdminPermission with parameter '*' and
    requesting DomainAdminPermission for a specific domain from a permission with parameter '*' will both return True.
    """
    def __init__(self, orgID):
        """Initialize domain admin permission.

        Parameters
        ----------
        orgID : int or '*'
            Organization ID this permission is for, or '*' to match all domains

        Raises
        ------
        ValueError
            `orgID` is neither an integer nor special identifier '*'
        """
        if orgID != "*" and not isinstance(orgID, int):
            raise ValueError("OrgAdminPermission parameter must be integer or '*'")
        self.__org = orgID

    def permits(self, permission):
        """Check if `permission` is represented.


        Returns
        -------
        bool
            True if `permission` is represented by this object, False otherwise
        """
        if isinstance(permission, OrgAdminPermission):
            return self._permits(permission)
        # `isinstance()` does not compare the inheritance of DAP and DA-RO-P properly
        if isinstance(permission, DomainAdminPermission) or isinstance(permission, DomainAdminROPermission):
            if permission.domainID == "*" or self.__org == "*":
                return True
            from orm.domains import Domains
            domainIDs = (d.ID for d in Domains.query.filter(Domains.orgID == self.__org).with_entities(Domains.ID).all())
            return permission.domainID in domainIDs
        return PermissionBase.permits(self, permission)

    def _permits(self, permission):
        """Check if permission is represented.

        Parameters
        ----------
        permission : OrgAdminPermission
            Permission to check for equality

        Returns
        -------
        bool
            True if domain IDs match or either is a wildcard organization, False otherwise
        """
        return "*" in (self.__org, permission.__org) or self.__org == permission.__org

    def __repr__(self):
        """Return string representation."""
        return "OrgAdminPermission({})".format(repr(self.__org))

    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Set containing "DomainAdmin" and "OrgAdmin" capabilities.
        """
        return {"DomainAdminRead", "DomainAdminWrite", "OrgAdmin"} | super().capabilities()

    @property
    def orgID(self):
        """Return domain parameter."""
        return self.__org

    @property
    def params(self):
        """Parameters relevant for the permission."""
        return self.__org


@Permissions.register("DomainPurge")
class DomainPurgePermission(PermissionBase):
    """Permission to purge domains.

    Does not grant permission to delete a domain on its own and is only effective if combined with an OrgAdmin permission.
    """
    @classmethod
    def capabilities(cls):
        """Get a set of capabilities provided by this permission.

        Returns
        -------
        set
            Set containing "DomainPurge" capability.
        """
        return {"DomainPurge"} | super().capabilities()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from flask import jsonify, request

import api
import idna

from api.core import API, secure
from api.security import loginUser, refreshToken, getSecurityContext, mkCSRF

from orm import DB
from services import Service
from tools import formats
from tools.tasq import TasQServer


@API.route(api.BaseRoute+"/status", methods=["GET"])
@secure(requireAuth=False)
def chkState():
    """Check status of the API."""
    return jsonify(message="API is operational",
                   database=DB is not None and DB.testConnection() is None,
                   tasq=TasQServer.running())


@API.route(api.BaseRoute+"/about", methods=["GET"])
@secure(requireAuth=False)
def getAbout(requireAuth=False):
    """Retrieve version information."""
    return jsonify(API=api.apiVersion, backend=api.backendVersion, schema=DB.version if DB is not None else None)


@API.route(api.BaseRoute+"/login", methods=["POST"])
@secure(requireAuth=False)
def login():
    if "user" not in request.form or "pass" not in request.form:
        refreshed = refreshToken()
        if refreshed is None:
            return jsonify(message="Incomplete login form"), 400
        return jsonify(grommunioAuthJwt=refreshed, csrf=mkCSRF(refreshed))
    success, val = loginUser(request.form["user"], request.form["pass"])
    if not success:
        API.logger.warning("Failed login attempt for user '{}' from '{}': {}"
                           .format(request.form["user"], request.remote_addr, val))
        return jsonify(message="Login failed", error=val), 401
    return jsonify(grommunioAuthJwt=val, csrf=mkCSRF(val))


@API.route(api.BaseRoute+"/profile", methods=["GET"])
@secure(authLevel="user")
def getProfile():
    user = request.auth["user"]
    capabilities = tuple(user.permissions().capabilities())
    return jsonify(user=user.fulldesc(exclude={"fetchmail"}), capabilities=capabilities)


def updatePasswordUnauth(data):
    from orm.users import Users
    user = Users.query.filter(Users.ID != 0, Users.username == data["user"]).first()
    if user is None:
        return jsonify(message="Invalid username or password"), 401
    if user.externID is not None:
        with Service("ldap") as ldap:
            error = ldap.authUser(user.externID, data["old"])
            if error is not None:
                return jsonify(message=error), 401
        return jsonify(message="Cannot modify LDAP imported user"), 400
    if not user.chkPw(data["old"]):
        return jsonify(message="Invalid username or password"), 401
    user.password = data["new"]
    DB.session.commit()
    return jsonify(message="Password updated for user '{}'".format(data["user"]))


@API.route(api.BaseRoute+"/passwd", methods=["PUT"])
@secure(requireAuth=False)
def updatePassword():
    data = request.get_json(silent=True)
    if data is None or "new" not in data or "old" not in data:
        return jsonify(message="Incomplete data"), 400
    if "user" in data:
        return updatePasswordUnauth(data)
    error = getSecurityContext("user")
    if error:
        return jsonify(message=error), 401
    user = request.auth["user"]
    if user.externID is not None:
        return jsonify(message="Cannot modify LDAP imported user"), 400
    if not user.chkPw(data["old"]):
        return jsonify(message="Old password does not match"), 403
    user.password = data["new"]
    DB.session.commit()
    return jsonify(message="Password updated")


def chkDomain(domain):
    try:
        domain = idna.encode(domain).decode("ascii")
        if not formats.domain.match(domain):
            return "Domain does not match required format"
        else:
            return None
    except idna.IDNAError as err:
        return err.args[0]
    except Exception as err:
        return "Unknown error ({})".format(type(err).__name__)


def chkEmail(email):
    if email.count("@") != 1:
        return "E-Mail address does not match required format"
    user, domain = email.split("@")
    err = chkDomain(domain)
    if err:
        return err
    email = user+"@"+idna.encode(domain).decode("ascii")
    if not formats.email.match(email):
        return "E-Mail address does not match required format"


@API.route(api.BaseRoute+"/chkFormat", methods=["GET"])
@secure(requireAuth=False)
def validateFormat():
    result = {}
    if "domain" in request.args:
        result["domain"] = chkDomain(request.args["domain"])
    if "email" in request.args:
        result["email"] = chkEmail(request.args["email"])
    return jsonify(result)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH


from cli import Cli, ArgumentParser


def _tryConnect(args, target):
    import requests
    cli = args._cli
    if args.verbose >= 1:
        cli.print("Trying '{}'...".format(target), end="")
    try:
        response = requests.get(target+"/api/v1/status", **args._cparm)
        if response.status_code != 200:
            if args.verbose >= 1:
                cli.print()
            return "Received unexpected response code {} from '{}', check if server is ok..."\
                .format(response.status_code, target)
        else:
            if args.verbose >= 1:
                cli.print(cli.col("success.", "green"))
            return 0
    except Exception as err:
        if args.verbose >= 1:
            cli.print(cli.col("connect failed ({})".format(type(err).__name__), "yellow"))
        return 1


def _getConnection(args):
    import re
    urlre = re.compile(r"^(?P<proto>https?://)?(?P<host>(([\w\.-]+)|(\[[:\da-fA-F]+\])))(:(?P<port>\d{1,5}))?$")
    url = urlre.match(args.host or "localhost")
    if url is None:
        return False, "Invalid host"
    host = url["host"]
    port = url["port"]
    if url["proto"] is None:
        for proto, defaultPort in (("https://", 8443), ("http://", 8080)):
            target = "{}{}:{}".format(proto, host, port or defaultPort)
            res = _tryConnect(args, target)
            if res == 0:
                port = port or defaultPort
                break
            elif res == 1:
                pass
            else:
                return False, res
        else:
            return False, "Failed to connect to server"
    else:
        proto = url["proto"]
        port = port or (8443 if proto == "https://" else 8080)
        target = "{}{}:{}".format(proto, host, port)
        res = _tryConnect(args, target)
        if res == 0:
            pass
        elif res == 1:
            return False, "Failed to connect to server"
        else:
            return False, res
    target = "{}{}:{}/api/v1/".format(proto, host, port)
    return True, (target, host, proto)


def _login(target, host, args):
    def debugout(*argv, **kwargs):
        if args.verbose >= 1:
            cli.print(*argv, **kwargs)

    import requests
    import urllib
    cli = args._cli
    user = args.user or "admin"
    if not args.password and host == "localhost" and not args.passwd:
        from api.security import mkJWT, mkCSRF
        debugout("Attempting passwordless login...", end="", flush=True)
        try:
            token = mkJWT({"usr": user})
            csrf = mkCSRF(token)
            code, response = _remoteExec(None, target, token, "version", cparm=args._cparm, csrf=csrf)
            if code == 200:
                debugout(cli.col("success.", "green"))
                return True, (token, csrf)
            debugout(cli.col("failed.", "yellow"))
        except Exception:
            debugout(cli.col("error.", "red"))
    try:
        passwd = args.passwd or cli.input("Password: ", secret=True)
        data = urllib.parse.urlencode({"user": user, "pass": passwd})
        response = requests.post(target+"login", data, headers={"Content-Type": "application/x-www-form-urlencoded"},
                                 **args._cparm)
        data = response.json() or {}
        if response.status_code != 200:
            return False, "{}{}".format(data.get("message", "Login failed"), ": "+data["error"] if "error" in data else "")
        if "grommunioAuthJwt" not in data:
            return False, "Login failed: invalid response"
        token = data["grommunioAuthJwt"]
        csrf = data.get("csrf", "")
        code, response = _remoteExec(None, target, token, "version", cparm=args._cparm, csrf=csrf)
        if code is None:
            return False, response
        if code == 404:
            return False, "Remote CLI not available on server"
        if code in (401, 403):
            return False, "Insufficient permissions to use remote CLI"
        if code != 200:
            if "message" in response:
                return False, "Remote execution failed: "+response["message"]
            return False, "Remote execution failed with code "+str(code)
        return True, (token, csrf)
    except Exception as err:
        return False, "Login failed ({})".format(type(err).__name__)


def _getPath(root, path):
    for part in path.split("."):
        root = root.get(part, {})
    return root if isinstance(root, str) else None


def _remoteExec(cli, target, token, command, mode="exec", redirectFs=False, cparm={}, csrf=""):
    import requests
    try:
        colored = cli.colored if cli is not None else False
        data = {"command": command, "mode": mode, "color": colored, "fs": {} if redirectFs else None}
        headers = {"Content-Type": "application/json", "X-Csrf-Token": csrf}
        response = requests.post(target+"system/cli", json=data, cookies={"grommunioAuthJwt": token}, headers=headers, **cparm)
        return response.status_code, response.json()
    except Exception as err:
        return None, "Remote execution failed ({})".format(type(err).__name__)


class RemoteCompleter:
    def __init__(self, cli, target, token):
        self.cli = cli
        self.target = target
        self.token = token
        self.cached = None
        self.completions = ()


    def rl_complete(self, text, state):
        """Provide completions from remote shell.

        Caches completions for the last query for consecutive calls with different states.

        Parameters
        ----------
        text : str
            Prefix to complete
        state : int
            Index of completion

        Returns
        -------
        str
            Completion according to state or None if completions are exhausted
        """
        if text != self.cached:
            code, data = _remoteExec(self.cli, self.target, self.token, text, mode="complete")
            self.completions = () if code != 200 or data is None or "completions" not in data else data["completions"]
            self.cached = text
        return None if state >= len(self.completions) else self.completions[state]


class RemoteCli(Cli):
    actionMap = {"discard": "d", "local": "s", "print": "V", "remote": "r"}

    def __init__(self, parent, target, token, host, redirectFs, autoSave, cparm, csrf=""):
        super().__init__("remote", fs=parent.fs, stdin=parent.stdin, stdout=parent.stdout, host=host, color=parent.colored)
        self.completer = RemoteCompleter(self, target, token)
        self.__target = target
        self.__token = token
        self.__csrf = csrf
        self.__parent = parent
        self.__cparm = cparm
        self.__redirectFs = redirectFs
        self.__autoSave = autoSave

    def _receiveFiles(self, fs):
        if fs is None:
            return
        for path, file in fs.items():
            mode = file.get("mode", "w")
            conf = file.get("conf")
            content = file.get("content", b"" if "b" in mode else "")
            while True:
                self.print("Received {}file {}".format("binary " if "b" in mode else "", self.col(path, attrs=["bold"])))
                if conf is not None:
                    from tools.config import Config
                    local = _getPath(Config, conf) or path
                else:
                    local = path
                if self.__autoSave is None or self.__autoSave not in self.actionMap:
                    choices = ["s", "d", "w"]
                    if local != path:
                        self.print("Remote path was '{}', but local path would be '{}' (according to {})"\
                                   .format(self.col(path, "red"), self.col(local, "yellow"), conf))
                        prompt = "{}, {}, (d)iscard, (v)iew, (w)rite to custom path"\
                                  .format(self.col("(s)ave to local path", "green"), self.col("save to (r)emote path", "yellow"))
                        choices.append("r")
                    else:
                        prompt = "(s)ave, (d)iscard, (w)rite to custom path"
                    if "b" not in mode:
                        prompt += ", (v)iew"
                        choices.append("v")
                    action = self.choice(prompt+"? [d]: ", choices, default="d")
                else:
                    action = self.actionMap[self.__autoSave]
                if action is None:
                    return
                if action == "s":
                    writePath = local
                elif action == "d":
                    self.print("File discarded")
                    break
                elif action == "w":
                    try:
                        writePath = self.input("Write file to: ")
                    except KeyboardInterrupt:
                        self.print("File discarded")
                        break
                elif action in ("v", "V"):
                    self.print(self.col(content, attrs=["bold"]))
                    if action == "V":
                        break
                    continue
                elif action == "r":
                    writePath = path
                if action in ("s", "w", "r"):
                    try:
                        with self.open(writePath, mode) as outfile:
                            outfile.write(file.get("content", ""))
                        self.print("File saved")
                        break
                    except Exception as err:
                        self.print(self.col("Failed to write file: "+" - ".join(str(arg) for arg in err.args), "yellow"))

    def _execute(self, command):
        code, data = _remoteExec(self, self.__target, self.__token, command, redirectFs=self.__redirectFs, cparm=self.__cparm,
                                 csrf=self.__csrf)
        if code is None:
            self.print(self.col(data, "red"))
            return 100
        if not isinstance(data, dict):
            data = {}
        if code != 200:
            self.print(self.col("Remote execution failed: "+data.get("message", "unknown error "+str(code)), "red"))
            return 101
        if "code" not in data or "stdout" not in data:
            self.print(self.col("The server returned an invalid response"))
            return 102
        self.print(data["stdout"], end="")
        self._receiveFiles(data.get("fs"))
        return data["code"]

    def execute(self, args, **kwargs):
        """Remote CLI execution override

        Forwards the command to the remote CLI and parses the response.

        Parameters
        ----------
        args : string or iterable
            Command line arguments, either as single string or split into individual parts
        **kwargs : any
            ignored.

        Returns
        -------
        int
            Execution return code
        """
        import shlex
        args = args if isinstance(args, str) else " ".join(shlex.quote(arg) for arg in args)
        return self._execute(args)

    def shell(self):
        """Alias for original Cli.execute(["shell"]).

        Returns
        -------
        int
            Return code of shell execution
        """
        return super().execute(["shell"])


def _cliRemoteSetupParser(subp: ArgumentParser):
    subp.add_argument("host", nargs="?", help="Host to connect to (default 'localhost')")
    subp.add_argument("user", nargs="?", help="User to connect with (default 'admin')")
    subp.add_argument("passwd", nargs="?", help="User password (default is to prompt)")
    subp.add_argument("--auto-save", choices=("local", "remote", "discard", "print"),
                      help="Automatically perform selected action when receiving files, instead of prompting")
    subp.add_argument("-c", "--command", help="Run command and exit (instead of starting shell)")
    subp.add_argument("--no-verify", action="store_true", help="Skip certificate verification")
    subp.add_argument("-p", "--password", action="store_true", help="Prompt for password even when connecting to localhost")
    subp.add_argument("--redirect-fs", action="store_true", help="Emulate CLI initiated read/write operations")
    subp.add_argument("-v", "--verbose", default=0, action="count", help="Print more information")


@Cli.command("connect", _cliRemoteSetupParser, help="Connect to a remote shell")
def cliRemote(args):
    cli = args._cli
    args._cparm = {}
    if args.no_verify:
        import urllib3
        import warnings
        warnings.filterwarnings("ignore", "", urllib3.exceptions.InsecureRequestWarning, "", 0)
        args._cparm["verify"] = False
    success, result = _getConnection(args)
    if not success:
        cli.print(cli.col(result, "red"))
        return 1
    target, host, proto = result
    if proto == "http://":
        cli.print(cli.col("Using insecure HTTP connection", "yellow"))
    success, result = _login(target, host, args)
    if not success:
        cli.print(cli.col(result, "red"))
        return 2
    token, csrf = result

    remoteCli = RemoteCli(cli, target, token, host, args.redirect_fs, args.auto_save, args._cparm, csrf)
    if args.command:
        return remoteCli.execute(args.command)
    return remoteCli.shell()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from flask import request, jsonify

import api
from api.core import API, secure
from api.security import checkPermissions

from tools import dbconf
from tools.permissions import SystemAdminPermission, SystemAdminROPermission

from orm import DB

@API.route(api.BaseRoute+"/system/dbconf/", methods=["GET"])
@secure(requireDB=True)
def getDbconfServices():
    checkPermissions(SystemAdminROPermission())
    from orm.misc import DBConf
    data = [entry[0] for entry in DBConf.query.with_entities(DBConf.service.distinct()).order_by(DBConf.service)]
    return jsonify(data=data)


@API.route(api.BaseRoute+"/system/dbconf/<service>/", methods=["GET"])
@secure(requireDB=True)
def getDbconfFiles(service):
    checkPermissions(SystemAdminROPermission())
    from orm.misc import DBConf
    data = [entry[0] for entry in DBConf.query.filter(DBConf.service == service)
                                              .with_entities(DBConf.file.distinct())
                                              .order_by(DBConf.file)]
    return jsonify(data=data)


@API.route(api.BaseRoute+"/system/dbconf/<service>/", methods=["PATCH"])
@secure(requireDB=True)
def renameDbconfService(service):
    checkPermissions(SystemAdminPermission())
    from orm.misc import DBConf
    data = request.get_json(silent=True)
    if data is None or "name" not in data:
        return jsonify(message="Missing or incomplete data"), 400
    count = DBConf.query.filter(DBConf.service == service).update({DBConf.service: data["name"]})
    if count == 0:
        DB.session.rollback()
        return jsonify(message="Service not found"), 404
    DB.session.commit()
    return jsonify(message="Success.")


@API.route(api.BaseRoute+"/system/dbconf/<service>/", methods=["DELETE"])
@secure(requireDB=True)
def deleteDbconfService(service):
    checkPermissions(SystemAdminPermission())
    from orm.misc import DBConf
    count = DBConf.query.filter(DBConf.service == service).delete()
    if count == 0:
        DB.session.rollback()
        return jsonify(message="Service not found"), 404
    DB.session.commit()
    return jsonify(message="Service deleted")


@API.route(api.BaseRoute+"/system/dbconf/<service>/<file>/", methods=["PUT", "PATCH"])
@secure(requireDB=True)
def updateDbconfFile(service, file):
    checkPermissions(SystemAdminPermission())
    from orm.misc import DBConf
    data = request.get_json(silent=True)
    if data is None:
        return jsonify(message="Missing data"), 400
    if "data" in data:
        conf = data["data"]
        if request.method == "PUT":
            DBConf.query.filter(DBConf.service == service, DBConf.file == file).delete()
        else:
            existing = DBConf.query.filter(DBConf.service == service, DBConf.file == file, DBConf.key.in_(conf)).all()
            for entry in existing:
                entry.value = conf.pop(entry.key)
        DB.session.add_all(DBConf(service=service, file=file, key=key, value=value) for key, value in conf.items())
    if "name" in data:
        count = DBConf.query.filter(DBConf.service == service, DBConf.file == file).update({DBConf.file: data["name"]})
        if count == 0:
            DB.session.rollback()
            return jsonify(message="File not found"), 404
    DB.session.commit()
    error = dbconf.commit(service, file)
    if error is None:
        return jsonify(message="Success.")
    return jsonify(message="Configuration updated but commit failed ({})".format(error))


@API.route(api.BaseRoute+"/system/dbconf/<service>/<file>/", methods=["DELETE"])
@secure(requireDB=True)
def deleteDbconfFile(service, file):
    checkPermissions(SystemAdminPermission())
    from orm.misc import DBConf
    count = DBConf.query.filter(DBConf.service == service, DBConf.file == file).delete()
    if count == 0:
        return jsonify(message="File not found"), 404
    DB.session.commit()
    return jsonify(message="File deleted")


@API.route(api.BaseRoute+"/system/dbconf/<service>/<file>/", methods=["GET"])
@secure(requireDB=True)
def getDbconfFile(service, file):
    checkPermissions(SystemAdminROPermission())
    from orm.misc import DBConf
    data = {entry.key: entry.value for entry in DBConf.query.filter(DBConf.service == service, DBConf.file == file)}
    return jsonify(data=data)


@API.route(api.BaseRoute+"/system/dbconf/commands", methods=["GET"])
@secure()
def getDbconfCommands():
    checkPermissions(SystemAdminROPermission())
    return jsonify(key=list(dbconf.keyMacros)+list(dbconf.keyCommits), file=list(dbconf.fileMacros)+list(dbconf.fileCommits),
                   service=list(dbconf.serviceMacros)+list(dbconf.serviceCommits))



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH


import logging
import threading
import queue


logger = logging.getLogger("tasq")


class Task:
    QUEUED = 0  # Stored in database
    LOADED = 1  # Loaded by a TaskQ server
    RUNNING = 2  # Currently not used
    COMPLETED = 3  # Completed successfully
    ERROR = 4  # Error during execution
    CANCELLED = 5  # Cancelled before execution

    _statnames = ("QUEUED", "LOADED", "RUNNING", "COMPLETED", "ERROR", "CANCELLED")

    @property
    def statename(self):
        return self._statnames[self.state] if 0 <= self.state < len(self._statnames) else "UNKNOWN"

    @property
    def done(self):
        """Whether the task is in completed state (either successful or not)."""
        return self.state >= self.COMPLETED

    def __init__(self, ID, command, params, state=LOADED, message=""):
        self.ID = ID
        self.command = command
        self.params = params
        self.state = state
        self.message = message

    def __repr__(self):
        return "<Task #{} ({}) '{}({})'>".format(self.ID, self.statename, self.command, self.params)


class Worker:
    def __init__(self, _queued=None, _finished=None):
        """Start TasQ worker.

        Omitting in- and output queues will not start the main loop,
        allowing to dispatch tasks manually.

        Parameters
        ----------
        _queued : Queue, optional
            Input queue. The default is None.
        _finished : TYPE, optional
            Output queue. The default is None.
        """
        self._queued, self._finished = _queued, _finished
        if None not in (_queued, _finished):
            self.run()
        self.__current = None

    def log(self, level, message):
        if self._finished is None:
            logger.log(logging.getLevelName(level), message)
        else:
            self._finished.put(Task(0, "control", dict(cmd="log", level=level, message=message)))

    def bump(self):
        if self._finished is not None:
            self._finished.put(Task(self.__current.ID, "control", dict(cmd="bump"), message=self.__current.message))

    def dispatch(self, task):
        from time import time
        func = self.cmap.get(task.command)
        if func is None:
            task.state = Task.ERROR
            task.message = "Unknown command '{}'".format(task.command)
        else:
            task.message = None
            try:
                start = time()
                func(self, task)
                duration = time()-start
            except Exception as err:
                import traceback
                from .config import Config
                self.log("ERROR", traceback.format_exc(Config["tasq"]["stackTraceDepth"]))
                task.state = Task.ERROR
                task.message = type(err).__name__+": "+" - ".join(str(arg) for arg in err.args)[:160]
        task.state = max(task.state, Task.COMPLETED)
        task.message = task.message or "Completed ({:.1f}ms)".format(1000*duration)
        return task

    def run(self):
        while True:
            self.__current = task = self._queued.get()
            self.dispatch(task)
            self._finished.put(task)
            self.__current = None

    def control(self, task):
        command = task.params.get("cmd")
        if command == "exit":
            raise SystemExit()
        else:
            raise Exception("Invalid or missing control command")

    def debug(self, task):
        command = task.params.get("cmd")
        if command == "bump":
            task.message = task.params.get("message", "Bump")
            self.bump()
            if "t" in task.params:
                import time
                time.wait(task.params["t"])
        elif command == "log":
            self.log(task.params.get("level", "INFO"), task.params.get("message", "(no message specified)"))
        elif command == "task":
            task.message = task.params.get("message", task.message)
            task.state = task.params.get("state", task.state)
        elif command == "wait":
            import time
            time.sleep(task.params.get("t", 5))
        else:
            raise Exception("Invalid or missing test command")

    def deleteFolder(self, task):
        if not {"homedir", "private", "folderID"}.issubset(task.params):
            raise Exception("Missing arguments for delFolder")
        from services import Service
        with Service("exmdb") as exmdb:
            host = task.params.get("homeserver") or exmdb.host
            client = exmdb.ExmdbQueries(host, exmdb.port, task.params["homedir"], task.params["private"])
            client.deleteFolder(task.params["homedir"], task.params["folderID"], task.params.get("clear", False))

    def _ldapSyncUser(self, user):
        from tools.ldap import downsyncObject
        result, code = downsyncObject(user)
        if code == 200:
            return {"ID": user.ID, "username": user.username, "code": 200, "message": "Synchronization successful"}
        return {"ID": user.ID, "username": user.username, "code": code, "message": result}

    def _ldapSyncImportUser(self, candidate, ldap, lang):
        from tools.ldap import importObject
        result, code = importObject(candidate, ldap, lang=lang)
        if code == 201:
            return dict(ID=result["ID"], username=result["username"], code=201, message="User created")
        return dict(username=candidate.email, code=code, message=result["message"])

    def _ldapSyncImportContact(self, candidate, ldap, orgID, domains):
        from tools.ldap import importContact
        synced, imported, failed = importContact(candidate, ldap, orgID, domains=domains)
        return [dict(ID=user.ID, username=user.username, code=201, message="Contact created") for user in imported] +\
               [dict(username=candidate.email, code=code, message=message) for message, code in failed]

    def _ldapSyncImport(self, ldap, orgID, domains, synced, lang, bump):
        syncStatus = []
        candidates = [candidate for candidate in ldap.searchUsers() if candidate.ID not in synced]
        domainnames = {domain.domainname for domain in domains}
        for candidate in candidates:
            bump()
            if candidate.type == "contact":
                syncStatus += self._ldapSyncImportContact(candidate, ldap, orgID, [domain.ID for domain in domains])
                continue
            if "@" not in candidate.email or (domainnames is not None and candidate.email.split("@", 1)[1] not in domainnames):
                syncStatus.append(dict(username=candidate.email, code=400, message="Invalid domain."))
                continue
            syncStatus.append(self._ldapSyncImportUser(candidate, ldap, lang))
        return syncStatus

    def _ldapSyncGroupMembers(self, orgID, ldap):
        from orm.users import Users
        from tools.ldap import syncGroupMembers
        self.message = "Synchronizing group members"
        self.bump()
        users = {user.externID for user in
                 Users.query.filter(Users.orgID == orgID, Users.externID != None).with_entities(Users.externID)}
        status = []
        for ldapgroup in ldap.searchUsers(types=("group",)):
            add, remove = syncGroupMembers(orgID, ldapgroup, ldap, users)
            if None in (add, remove):
                status.append(dict(username=ldapgroup.email, code=404, message="Group not found"))
            else:
                status.append(dict(username=ldapgroup.email, code=200,
                                   message=f"{add} added to/{remove} removed from member list"
                                   if add+remove else "member list unchanged"))
        return status

    def ldapSync(self, task):
        def bump():
            nonlocal last
            if time.time()-last < updateInterval:
                return
            updateMessage()
            last = time.time()
            self.bump()

        def statusCat(code):
            return "created" if code == 201 else "synced" if code == 200 else "error"

        def updateMessage():
            task.message = "{}/{} synced".format(counts["synced"], counts["sync"])
            if counts["created"]:
                task.message += ", {} created".format(counts["created"])
            if counts["error"]:
                task.message += ", {} error{}".format(counts["error"], "" if counts["error"] == 1 else "s")

        from orm import DB
        from orm.domains import Domains, OrgParam, Orgs
        from orm.users import Aliases, Users
        from services import Service, ServiceUnavailableError
        import time

        DB.session.rollback()
        start = last = time.time()
        orgID = task.params.get("orgID")
        domainID = task.params.get("domainID")
        updateInterval = task.params.get("updateInterval", 5)
        Aliases.NTactive(False)
        Users.NTactive(False)

        domainFilter = ()
        noLdapOrgs = ()  # IDs of orgs without LDAP config override
        if domainID is not None:
            domainFilter = (Domains.ID == domainID,)
            domain = Domains.query.filter(Domains.ID == domainID).with_entities(Domains.orgID).first()
            orgIDs = [domain.orgID]
            userfilter = [Users.domainID == domainID]
        elif orgID is not None:
            orgIDs = [orgID]
            userfilter = [Users.orgID == orgID]
        else:
            noLdapOrgs = {org.ID for org in Orgs.query.with_entities(Orgs.ID)}
            ldapOrgs = set(OrgParam.ldapOrgs())
            orgIDs = ldapOrgs.union((0,))
            noLdapOrgs -= ldapOrgs
            userfilter = ()

        users = Users.query.filter(Users.externID != None, *userfilter).all()
        counts = dict(created=0, synced=0, error=0, create=0, sync=len(users))
        syncStatus = []
        synced = set()

        for user in users:
            bump()
            try:
                status = self._ldapSyncUser(user)
                syncStatus.append(status)
                counts[statusCat(status["code"])] += 1
                if status["code"] == 200:
                    synced.add(user.externID)
            except ServiceUnavailableError as err:
                syncStatus.append(dict(ID=user.ID, username=user.username, code=503, message=err.args[0]))
                counts["error"] += 1

        if task.params.get("import"):
            for orgID in orgIDs:
                domains = Domains.query.filter(Domains.orgID == orgID, *domainFilter)\
                                       .with_entities(Domains.ID, Domains.domainname).all()
                if orgID == 0 and noLdapOrgs:
                    domains += Domains.query.filter(Domains.orgID.in_(noLdapOrgs))\
                                      .with_entities(Domains.ID, Domains.domainname).all()
                ldap = Service("ldap", orgID).service()
                try:
                    status = self._ldapSyncImport(ldap, orgID, domains, synced, task.params.get("lang"), bump)
                    counts["synced"] += sum(1 for s in status if s["code"] == 200)
                    counts["created"] += sum(1 for s in status if s["code"] == 201)
                    counts["error"] += sum(1 for s in status if s["code"] not in (200, 201))
                    syncStatus += status
                    self._ldapSyncGroupMembers(orgID, ldap)
                except ServiceUnavailableError:
                    pass

        Aliases.NTactive(True)
        Users.NTactive(True)


        updateMessage()
        task.message += " ({:.1f}s)".format(time.time()-start)
        task.params["result"] = syncStatus

    cmap = {"control": control, "debug": debug, "delFolder": deleteFolder, "ldapSync": ldapSync}


class TasQServer:
    STOPPED = 0
    STARTING = 1
    STARTED = 2
    STOPPING = 3

    _queued = queue.Queue()
    _finished = queue.Queue()
    _state = STOPPED
    _clerk = None
    _online = True
    _active = {}
    _active_lock = threading.Lock()
    _localID = 0
    _workers = []

    @classmethod
    def _schedule(cls, task):
        with cls._active_lock:
            cls._active[task.ID] = (task, threading.Condition(cls._active_lock))
        cls._queued.put(task)
        return task

    @classmethod
    def create(cls, command, params, synced=True, permission=None, inline=None):
        """Create a new task.

        Parameters
        command : str
            Name of the command
        params : dict
            Command specific parameters
        synced : bool, optional
            Whether to synchronize the task with the database. The default is True.
        permission : PermissionBase, optional
            Restrict access to users with permission. The default is None.
        inline : bool, optional
            Do not execute async, but dispatch in current thread.
            If set to None, only execute inline if TasQ server is not running.
            The default is None.

        Raises
        ------
        ValueError
            Trying to create a "control" command.

        Returns
        -------
        Task
            The created task or None if synced is True and the server is not running.
        """
        if command == "control":
            raise ValueError("Cannot create control commands")
        if inline or (inline is None and not cls.running()):
            return Worker().dispatch(Task(0, command, params))
        elif cls.online() and synced:
            from orm.misc import DB, TasQ
            dbtask = TasQ(dict(command=command, params=params))
            dbtask.state = Task.LOADED if cls.running() else Task.QUEUED
            dbtask.permission = permission
            DB.session.add(dbtask)
            DB.session.commit()
            if cls.running():
                return cls._schedule(Task(dbtask.ID, command, params))
            else:
                return Task(dbtask.ID, command, params)
        else:
            if not cls.running():
                logger.warning("Added local task but TasQ server is not running")
            cls._localID -= 1
            return cls._schedule(Task(cls._localID, command, params))

    @classmethod
    def start(cls, workers=None, online=True):
        """Start the TasQ server.

        Has no effect if the server is already running.

        If procs is None, the number of workers is defined by the
        configuration (default 1).

        Parameters
        ----------
        procs : int, optional
            Number of workers to start. The default is None.
        online : bool, optional
            Whether to run in online mode (tasks are synchronized with the database).
            The default is True.
        """
        if cls._state != cls.STOPPED:
            return
        cls._state = cls.STARTING
        import atexit
        from .config import Config
        atexit.register(cls.stop)
        conf = Config.get("tasq", {})
        workers = workers or conf.get("workers", 1)
        logger.info("Starting TasQ server with {} worker{}".format(workers, "" if workers == 1 else "s"))
        cls._workers = [threading.Thread(target=Worker, args=(cls._queued, cls._finished), name="TasQ Worker")
                        for _ in range(workers)]
        for worker in cls._workers:
            worker.start()
            logger.debug("Started worker with id "+str(worker.ident))
        cls._clerk = threading.Thread(target=cls._process)
        cls._clerk.start()
        cls._online = online
        cls.pull()
        cls._state = cls.STARTED

    @classmethod
    def stop(cls, timeout=None):
        """Stop the TasQ server.

        Parameters
        ----------
        timeout : float, optional
            Maximum number of seconds to wait for worker processes to exit. The default is None.
        """
        if not cls.running():
            return
        cls._state = cls.STOPPING
        from datetime import datetime
        from time import time
        timeout = timeout+time() if timeout is not None else None
        logger.info("Shutting down TasQ server")
        cancelled = []
        try:
            while True:
                cancelled.append(cls._queued.get(False))
        except queue.Empty:
            pass
        for proc in cls._workers:
            cls._queued.put(Task(0, "control", {"cmd": "exit", "dbg": "proc"}))
        with cls._active_lock:
            for task in cancelled:
                if task.ID > 0:
                    from orm.misc import DB, TasQ
                    dbtask = TasQ.query.filter(TasQ.ID == task.ID).first() or \
                             TasQ(dict(command=task.command, params=task.params))
                    dbtask.state = Task.QUEUED
                    dbtask.message = "Restored on TasQ server shutdown"
                    dbtask.updated = datetime.now()
                    DB.session.commit()
                else:
                    task.state = Task.CANCELLED
                    task.message = "TasQ server in offline mode was shut down before task completed"
                tracker = cls._active.pop(task.ID, None)
                if tracker is None:
                    continue
                tracker[0].state = task.state
                tracker[1].message = task.message
                tracker[1].notify_all()
        if len(cancelled):
            logger.info("Putting {} loaded task{} back into the database"
                        .format(len(cancelled), "" if len(cancelled) == 1 else "s"))
        logger.debug("Waiting for workers to exit")
        for proc in cls._workers:
            proc.join(max(timeout-time(), 0) if timeout is not None else None)
        cls._finished.put(Task(0, "control", {"cmd": "exit", "dbg": "thread"}))
        cls._clerk.join()
        cls._state = cls.STOPPED
        logger.info("TasQ server stopped")

    @classmethod
    def online(cls):
        """Try to enable online mode.

        Has no effect if online mode is already enabled or explicitly disabled.

        Returns
        -------
        bool
            Whether the server is now in online mode
        """
        if cls._online is not None:
            return cls._online
        cls._online = True
        return cls.pull() is not None

    @classmethod
    def pull(cls):
        """Import queued tasks from the database.

        Only has an effect if the server is running and in online mode.
        """
        if not cls.running() or not cls._online:
            return 0
        from orm import DB
        from datetime import datetime
        if DB is None or not DB.minVersion(102):
            cls._online = None
            msg = "Database unavailable" if DB is None else "Schema version too old (n102 required)"
            logger.warning(msg + " - falling back to offline mode.")
            return None
        from orm.misc import TasQ
        waiting = TasQ.query.filter(TasQ.state == Task.QUEUED).with_for_update().all()
        for w in waiting:
            if w.command == "control":
                w.state = Task.CANCELLED
                w.message = "Task dropped during import: invalid command"
            else:
                w.state = Task.LOADED
                w.message = "Imported task from database"
            w.updated = datetime.now()
        tasks = [(Task(w.ID, w.command, w.params), w) for w in waiting if w.command != "control"]
        DB.session.commit()
        for task, dbtask in tasks:
            cls._schedule(task)
        logger.info("Pulled {} task{} from database".format(len(tasks), "" if len(tasks) == 1 else "s"))
        return len(tasks)

    @classmethod
    def running(cls):
        """Check if the TasQ server is currently running.

        Returns
        -------
        bool
            Whether the server is running.
        """
        return cls._state in (cls.STARTED, cls.STARTING)

    @classmethod
    def queued(cls):
        """Return number of tasks waiting to be processed."""
        return cls._queued.qsize()

    @classmethod
    def workers(cls):
        """Return number of active worker processes."""
        return sum(1 for proc in cls._workers if proc.is_alive())

    @classmethod
    def wait(cls, taskID, timeout=None):
        """Wait for a task to finish.

        Returns immediately if the is not active.

        The task ID can be retrieved from the task object returned by `create`.

        Parameters
        ----------
        taskID : int
            ID of the task
        timeout : float, optional
            Maximum time (in seconds) to wait for task completion. The default is None.
        """
        with cls._active_lock:
            tracker = cls._active.get(taskID)
            if tracker is not None:
                tracker[1].wait(timeout)

    @classmethod
    def _process(cls):
        logger.debug("Clerk started")
        from datetime import datetime
        while True:
            task = cls._finished.get()
            if task.command == "control":
                if not task.params:
                    continue
                cmd = task.params.get("cmd")
                if cmd == "exit":
                    logger.debug("Clerk stopped")
                    return
                elif cls._online and cmd == "bump":
                    from orm.misc import DB, TasQ
                    dbtask = TasQ.query.filter(TasQ.ID == task.ID).first()
                    if dbtask is not None:
                        dbtask.message = task.message or ""
                        dbtask.updated = datetime.now()
                        DB.session.commit()
                elif cmd == "log":
                    try:
                        logger.log(logging.getLevelName(task.params.get("level", "INFO")),
                                   "<worker> "+task.params.get("message", "(no message specified)"))
                    except Exception:
                        pass
                continue
            with cls._active_lock:
                tracker = cls._active.pop(task.ID, None)
                if cls._online:
                    from orm.misc import DB, TasQ
                    dbtask = TasQ.query.filter(TasQ.ID == task.ID).first()
                    if dbtask is not None:
                        dbtask.state = task.state
                        dbtask.message = task.message
                        dbtask.updated = datetime.now()
                        dbtask.params = task.params
                    DB.session.commit()
                if tracker is not None:
                    tracker[0].state = task.state
                    tracker[0].message = task.message
                    tracker[1].notify_all()
            logger.debug("Task #{} completed ({})".format(task.ID, task.statename))

    class mktask:
        @staticmethod
        def deleteFolder(homedir, folderID, private, clear=False, permission=None, homeserver=None):
            return TasQServer.create("delFolder", dict(homedir=homedir, folderID=folderID, private=private, clear=clear,
                                                       homeserver=homeserver.hostname if homeserver else None),
                                     permission=permission)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

import json

from sqlalchemy import Column, ForeignKey
from sqlalchemy.dialects.mysql import INTEGER, TEXT, VARCHAR
from sqlalchemy.orm import relationship

from tools.DataModel import DataModel, Id, Int, RefProp, Text

from . import DB


class AdminRoles(DataModel, DB.Base):
    __tablename__ = "admin_roles"

    ID = Column("id", INTEGER(10, unsigned=True), unique=True, primary_key=True)
    name = Column("name", VARCHAR(32), unique=True, nullable=False)
    description = Column("description", VARCHAR(256))

    permissions = relationship("AdminRolePermissionRelation", cascade="all, delete-orphan", single_parent=True)
    users = relationship("AdminUserRoleRelation", cascade="all, delete-orphan", back_populates="role")

    _dictmapping_ = ((Id(), Text("name", flags="patch")),
                     (Text("description", flags="patch"),),
                     (RefProp("permissions", flags="patch, managed", link="ID"),
                      RefProp("users", link="userID", flat="user", flags="patch")))


class AdminRolePermissionRelation(DataModel, DB.Base):
    __tablename__ = "admin_role_permission_relation"

    ID = Column("id", INTEGER(10, unsigned=True), primary_key=True)
    roleID = Column("role_id", INTEGER(10, unsigned=True), ForeignKey(AdminRoles.ID), nullable=False)
    permission = Column("permission", VARCHAR(64), nullable=False)
    _params = Column("parameters", TEXT)

    role = relationship(AdminRoles, back_populates="permissions")

    _dictmapping_ = ((Id(), Text("permission", flags="init"), {"attr": "params", "flags": "patch"}), (), (Int("roleID"),))

    @property
    def params(self):
        return json.loads(self._params) if self._params is not None else None

    @params.setter
    def params(self, value):
        self._params = json.dumps(value, separators=(',', ':'))

    def __init__(self, props, role, *args, **kwargs):
        self.role = role
        self.fromdict(props, *args, **kwargs)

    def fromdict(self, patches, *args, **kwargs):
        if "permission" in patches:
            from tools.permissions import Permissions
            try:
                Permissions.create(patches["permission"], patches.get("params"))
            except KeyError as err:
                raise ValueError(*err.args)
        return DataModel.fromdict(self, patches, *args, **kwargs)


class AdminUserRoleRelation(DataModel, DB.Base):
    __tablename__ = "admin_user_role_relation"

    userID = Column("user_id", INTEGER(10, unsigned=True), ForeignKey("users.id", ondelete="cascade"), primary_key=True)
    roleID = Column("role_id", INTEGER(10, unsigned=True), ForeignKey(AdminRoles.ID), primary_key=True)

    user = relationship("Users", viewonly=True)
    role = relationship(AdminRoles, back_populates="users")

    _dictmapping_ = ((RefProp("user"),),
                     (RefProp("role"),))

    def __init__(self, userID, role, *args, **kwargs):
        self.userID = userID
        if isinstance(role, int):
            self.roleID = role
        else:
            self.role = role

from .users import Users



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

import yaml
import logging
from os import scandir

logger = logging.getLogger("config")


def _defaultConfig():
    _defaultSyncPolicy = {
      "allowbluetooth": 2,
      "allowbrowser": 1,
      "allowcam": 1,
      "allowconsumeremail": 1,
      "allowdesktopsync": 1,
      "allowhtmlemail": 1,
      "allowinternetsharing": 1,
      "allowirda": 1,
      "allowpopimapemail": 1,
      "allowremotedesk": 1,
      "allowsimpledevpw": 1,
      "allowsmimeencalgneg": 2,
      "allowsmimesoftcerts": 1,
      "allowstoragecard": 1,
      "allowtextmessaging": 1,
      "allowunsignedapps": 1,
      "allowunsigninstallpacks": 1,
      "allowwifi": 1,
      "alphanumpwreq": 0,
      "approvedapplist": [],
      "attenabled": 1,
      "devencenabled": 0,
      "devpwenabled": 0,
      "devpwexpiration": 0,
      "devpwhistory": 0,
      "maxattsize": "",
      "maxcalagefilter": 0,
      "maxdevpwfailedattempts": 8,
      "maxemailagefilter": 0,
      "maxemailbodytruncsize": -1,
      "maxemailhtmlbodytruncsize": -1,
      "maxinacttimedevlock": 900,
      "mindevcomplexchars": 3,
      "mindevpwlenngth": 4,
      "pwrecoveryenabled": 0,
      "reqdevenc": 0,
      "reqencsmimealgorithm": 0,
      "reqencsmimemessages": 0,
      "reqmansyncroam": 0,
      "reqsignedsmimealgorithm": 0,
      "reqsignedsmimemessages": 0,
      "reqstoragecardenc": 0,
      "unapprovedinromapplist": []
    }
    return {
        "DB": {
            "sessionTimout": 28800,
            },
        "dns": {
            "dudIP": "172.16.254.254",
            "externalResolvers": ["1.1.1.1", "1.0.0.1"]
            },
        "openapi": {
            "validateRequest": True,
            "validateResponse": True
            },
        "options": {
            "antispamUrl": "http://localhost:11334",
            "disableDB": False,
            "dataPath": "/usr/local/share/grommunio-admin-common",
            "portrait": "portrait.jpg",
            "domainStoreRatio": 10,
            "domainPrefix": "/var/db/gromox/domain/",
            "userPrefix": "/var/db/gromox/user/",
            "exmdbHost": "::1",
            "exmdbPort": "5000",
            "domainStorageLevels": 1,
            "userStorageLevels": 2,
            "domainAcceleratedStorage": None,
            "userAcceleratedStorage": None,
            "dashboard": {
                "services": []
                },
            "serverPolicy": "round-robin",
            "updateLogPath": "/var/log/grommunio-update.log",
            "updateSkriptPath": "/usr/sbin/grommunio-update",
            },
        "security": {
            "jwtPrivateKeyFile": "/var/db/grommunio-admin-api/auth-private.pem",
            "jwtPublicKeyFile": "/var/db/grommunio-admin-api/auth-public.pem",
            "rsaKeySize": 4096,
            },
        "mconf": {
          "ldapPath": "/etc/gromox/ldap_adaptor.cfg",
          "authmgrPath": "/etc/gromox/authmgr.cfg"
          },
        "logs": {},
        "sync": {
            "syncStateFolder": "GS-SyncState",
            "defaultPolicy": _defaultSyncPolicy,
            "policyHosts": ["127.0.0.1", "localhost", "::1", "::ffff:127.0.0.1"]
            },
        "chat": {
            "connection": {},
            },
        "tasq": {
            "stackTraceDepth": 7,
            },
        }


def initLoggers():
    if "logging" not in Config:
        return
    logconf = Config["logging"]
    logging.getLogger().setLevel(logconf.get("level", logging.WARNING))
    for logger, conf in logconf.get("loggers", {}).items():
        logging.getLogger(logger).setLevel(conf.get("level", logging.NOTSET))


def _recursiveMerge_(dst, add):
    """Recursively merge two dictionaries.

    Add values from `src` to `dst`. If a key from `src` is already present in `dst`,
    the merge strategy depends on their types:
        - If both are lists, the lists are concatenated
        - If both are dicts, they are merged recursively
        - Otherwise the value from `dst` is overwritten
    """
    assert type(dst) is dict and type(add) is dict
    for key in add.keys():
        if key in dst:
            if type(dst[key]) is list and type(add[key]) is list:
                dst[key] += add[key]
            elif type(dst[key]) is dict and type(add[key]) is dict:
                _recursiveMerge_(dst[key], add[key])
            else:
                dst[key] = add[key]
        else:
            dst[key] = add[key]


def _loadConfig_():
    """Load configuration file.

    Try to load configuration from './config.yaml'.
    If the file exists, the default configuration is updated.

    If the optional value 'confdir' is present,
    the specified directory is searched for further YAML files,
    which are recursively merged into the config.
    """
    config = _defaultConfig()
    try:
        with open("config.yaml", "r", encoding="utf-8") as file:
            _recursiveMerge_(config, yaml.load(file, Loader=yaml.SafeLoader))
    except Exception as err:
        logger.error("Failed to load 'config.yaml': {}".format(" - ".join(str(arg) for arg in err.args)))
    if "confdir" in config:
        try:
            configFiles = sorted([file.path for file in scandir(config["confdir"]) if file.name.endswith(".yaml")])
        except Exception as err:
            logger.error("Failed to stat '{}': ".format(config["confdir"])+" - ".join(str(arg) for arg in err.args))
            configFiles = ()
        for configFile in configFiles:
            try:
                with open(configFile, encoding="utf-8") as file:
                    confd = yaml.load(file, Loader=yaml.SafeLoader)
                if confd is not None:
                    _recursiveMerge_(config, confd)
            except Exception as err:
                logger.error("Failed to load '{}': {}".format(configFile, " - ".join(str(arg) for arg in err.args)))
    return config


Config = _loadConfig_()


def validate():
    """Verify configuration validity.

    Returns
    -------
    str
        Error message, or None if validation succeeds
    """
    import openapi_spec_validator
    from openapi_schema_validator import OAS30Validator
    version = [int(part) for part in openapi_spec_validator.__version__.split(".")]
    if version < [0, 5, 0]:
        from openapi_spec_validator.exceptions import ValidationError
    else:
        from openapi_spec_validator.validation.exceptions import ValidationError
    try:
        with open("res/config.yaml", encoding="utf-8") as file:
            configSchema = yaml.load(file, yaml.loader.SafeLoader)
    except Exception:
        return "Could not open schema file"
    validator = OAS30Validator(configSchema)
    try:
        validator.validate(Config)
    except ValidationError as err:
        return err.args[0]



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from .misc import createMapping

import os
import shlex

keyCommits = {"postconf -e $ENTRY"}
fileCommits = {"sudo postconf -e $FILE_S && systemctl reload postfix"}
serviceCommits = {"systemctl reload $SERVICE",
                  "systemctl restart $SERVICE"}

keyMacros = {"#POSTCONF": "postconf -e $ENTRY"}
fileMacros = {"#POSTCONF": "sudo postconf -e $FILE_S && systemctl reload postfix"}
serviceMacros = {"#RELOAD": "systemctl reload $SERVICE",
                 "#RESTART": "systemctl restart $SERVICE",
                 }


def subVars(command, data):
    out = ""
    index = command.find("$")
    last = 0
    while index != -1:
        out += command[last:index]
        if index == len(command)-1:
            return out+"$"
        if command[index+1] == "$":
            last = index+2
            out += "$"
        else:
            last = index+1
            while last != len(command) and (command[last].isalnum() or command[last] == '_'):
                last += 1
            token = command[index+1:last]
            content = data.get(token)
            if not content:
                continue
            if isinstance(content, tuple):
                content = content[0].join(shlex.quote(arg) for arg in content[1])
            else:
                content = shlex.quote(content)
            out += content
        index = command.find("$", last)
    out += command[last:]
    return out


def _commitKey(exprmap, service, file, key):
    command = exprmap["commit_key"]
    command = keyMacros.get(command, command)
    if command not in keyCommits:
        return "Command not allowed for key trigger"
    if "$" in command:
        from orm.misc import DBConf
        entry = DBConf.query.filter(DBConf.service == service, DBConf.file == file, DBConf.key == key)\
                            .with_entities(DBConf.value).first()
        data = dict(KEY=key, VALUE=entry.value, ENTRY=key+"="+entry.value, FILENAME=file, SERVICE=service)
        command = subVars(command, data)
    ret = os.system(command)
    if ret:
        return "Command exited with status code "+str(ret)


def _commitFile(exprmap, service, file):
    command = exprmap["commit_file"]
    command = fileMacros.get(command, command)
    if command not in fileCommits:
        return "Command not allowed for file trigger"
    if "$" in command:
        from orm.misc import DBConf
        entries = DBConf.query.filter(DBConf.service == service, DBConf.file == file)\
                              .with_entities(DBConf.key, DBConf.value).all()
        filedata = [entry.key+"="+entry.value for entry in entries]
        data = dict(FILE=("\n", filedata), FILE_S=(" ", filedata), FILENAME=file, SERVICE=service)
        command = subVars(command, data)
    ret = os.system(command)
    if ret:
        return "Command exited with status code "+str(ret)


def _commitService(exprmap, service):
    command = exprmap.get("commit_service")
    command = serviceMacros.get(command, command)
    if command is not None:
        if command not in serviceCommits:
            return "Command not allowed for service trigger"
        command = subVars(command, {"SERVICE": service})
        ret = os.system(command)
        if ret:
            return "Command exited with status code "+str(ret)


def commit(service, file=None, key=None):
    """Commit configuration changes.

    Parameters
    ----------
    service : str
        Name of the service
    file : str, optional
        Name of the file or None to commit the service
    key : str, optional
        Name of the key or None to commit the file

    Returns
    -------
    str
        Error message or None if successful
    """
    from orm.misc import DBConf
    exprs = DBConf.query.filter(DBConf.service == "grommunio-dbconf", DBConf.file == service, DBConf.key.like("commit_%"))\
                        .with_entities(DBConf.key, DBConf.value).all()
    if len(exprs) == 0:
        return None
    exprmap = createMapping(exprs, lambda x: x.key, lambda x: x.value)
    if key is not None and "commit_key" in exprmap:
        return _commitKey(exprmap, service, file, key)
    if file is not None and "commit_file" in exprmap:
        return _commitFile(exprmap, service, file)
    return _commitService(exprmap, service)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH

from flask import request
import hashlib
import logging
import time
import jwt

from base64 import b64encode
from services import Service
from tools.config import Config


logger = logging.getLogger("security")
_priFile = Config["security"]["jwtPrivateKeyFile"]
_pubFile = Config["security"]["jwtPublicKeyFile"]
try:
    with open(_priFile, "rb") as file:  # Private key for JWT signing
        jwtPrivkey = file.read()
    with open(_pubFile, "rb") as file:  # Public key for JWT signature verification
        jwtPubkey = file.read()
except Exception:
    import logging
    logger.warn("Could not load JWT RSA keys ('{}', '{}'), generating new...".format(_priFile, _pubFile))
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import asymmetric, serialization as out
    import os
    jwtPrivkey = asymmetric.rsa.generate_private_key(65537, Config["security"]["rsaKeySize"], default_backend())
    jwtPubkey = jwtPrivkey.public_key()
    try:
        with open(_priFile, "wb") as priFile, open(_pubFile, "wb") as pubFile:
            priFile.write(jwtPrivkey.private_bytes(out.Encoding.PEM, out.PrivateFormat.TraditionalOpenSSL, out.NoEncryption()))
            pubFile.write(jwtPubkey.public_bytes(out.Encoding.PEM, out.PublicFormat.PKCS1))
        os.chmod(_priFile, 0o600)
    except Exception as err:
        logger.info("An exception occurred ("+type(err).__name__+"): "+" - ".join(str(arg) for arg in err.args))
        logger.error("Failed to save JWT RSA keys, logins will not persist across API restarts")


def getUser():
    """Load currently logged in user from database.

    Returns
    -------
    str
        Error message or None if successful.
    """
    if not hasattr(request, "auth"):
        return "Not authenticated"
    if "user" in request.auth:
        return
    try:
        from orm.users import Users
        user = Users.query.filter(Users.username == request.auth["claims"]["usr"]).first()
    except Exception as err:
        logger.warn("getUser() failed ({}): {}".format(type(err).__name__, " - ".join(str(arg) for arg in err.args)))
        return "Failed to get user information from database"
    if user is None:
        return "Invalid user"
    request.auth["user"] = user


def getSecurityContext(authLevel, validateCSRF=None):
    """Create security context for the request.

    Check for `jwt` cookie in the request and try to decode it. If token is valid, the claims are saved in `request.auth`

    If authentication is disabled in the configuration (security.requireAuth=False), the default security context is used
    (security.defaultContext).

    Returns
    -------
    str
        Error message or None if successful

    """
    request.auth = {}
    cookie = request.cookies.get("grommunioAuthJwt")
    if cookie is None:
        return "No token provided"
    success, val = checkToken(cookie)
    if not success:
        return val
    validateCSRF = not request.method == "GET" if validateCSRF is None else validateCSRF
    if validateCSRF and not checkCSRF(cookie):
        return "Invalid or missing CSRF token"
    request.auth["claims"] = val
    if authLevel == "user":
        return getUser()


def mkCSRF(token):
    """Generate CSRF token from JWT."""
    return b64encode(hashlib.sha3_256(token.encode("ascii")).digest()).decode("ascii")


def mkJWT(claims):
    """Generate signed JWT.

    `exp` field is automatically added if not present. Token expiration can be configured by security.jwtExpiresAfter
    (in seconds), default is one week.

    Parameters
    ----------
    claims : dict
        Claims included in the JWT

    Returns
    -------
    bytes
        Signed JWT token
    """
    from tools.config import Config
    if "exp" not in claims:
        claims["exp"] = int(time.mktime(time.gmtime())+Config["options"].get("jwtExpiresAfter", 7*24*60*60))
    token = jwt.encode(claims, jwtPrivkey, "RS256")
    return token.decode("ascii") if isinstance(token, bytes) else token


def checkCSRF(token):
    """Check validity of CSRF token.

    Parameters
    ----------
    token : str
        JWT

    Returns
    -------
    bool
        True if a valid CSRF token is present, False otherwise

    """
    return request.headers.get("X-Csrf-Token") == mkCSRF(token)


def checkToken(token):
    """Check jwt validity.

    Parameters
    ----------
    token : str
        JWT to check

    Returns
    -------
    bool
        True if valid, false otherwise
    dict / str
        Dict containing the JWT claims if successful, error message otherwise
    """
    try:
        claims = jwt.decode(token, jwtPubkey, algorithms=["RS256"])
    except jwt.ExpiredSignatureError:
        return False, "Token has expired"
    except jwt.InvalidSignatureError:
        return False, "Invalid token signature"
    except Exception:
        return False, "invalid token"
    return True, claims


def userLoginAllowed(user):
    from orm.roles import AdminUserRoleRelation
    return user.ID == 0 or (AdminUserRoleRelation.query.filter(AdminUserRoleRelation.userID == user.ID).count() != 0 and
                            user.addressStatus == 0)


def refreshToken():
    """Refresh user token.

    Check if the current token is valid and if the current user still is allowed to login.
    """
    from orm.users import Users
    if "grommunioAuthJwt" not in request.cookies:
        return
    success, claims = checkToken(request.cookies["grommunioAuthJwt"])
    if not success:
        return
    user = Users.query.filter(Users.username == claims["usr"]).first()
    if not user or not userLoginAllowed(user):
        return
    if "exp" in claims:
        claims.pop("exp")
    return mkJWT(claims)


def loginUser(username, password):
    """Try to authenticate user.

    Parameters
    ----------
    username : str
        User e-mail address
    password : str
        Password to check

    Returns
    -------
    bool
        Whether authentication was successful
    str
        JWT if successful, error message otherwise.
    """
    from orm.users import Users, Altnames
    user: Users = Users.query.join(Altnames, isouter=True)\
                             .filter((Users.username == username) | (Altnames.altname == username)).first()
    if user is None:
        return False, "Invalid username or password"
    if user.externID is not None:
        with Service("ldap", user.orgID) as ldap:
            error = ldap.authUser(user.externID, password)
            if error:
                return False, error
    elif not user.chkPw(password):
        return False, "Invalid username or password"
    if not userLoginAllowed(user):
        return False, "Access denied"
    try:
        token = mkJWT({"usr": user.username})
    except Exception:
        return False, "Token generation failed"
    return True, token


def checkPermissions(*requested):
    """Check if current user has requested permissions.

    Parameters
    ----------
    *requested : PermissionBase
        Permissions the user needs

    Raises
    ------
    InsufficientPermissions
        At least one permission check failed

    Returns
    -------
    None.
    """
    from .errors import InsufficientPermissions
    error = getUser()
    if error is not None:
        raise InsufficientPermissions(error)
    user = request.auth["user"]
    if user.ID != 0 and user.addressStatus != 0:
        raise InsufficientPermissions("Account deactivated")
    permissions = user.permissions()
    if not all(permission in permissions for permission in requested):
        raise InsufficientPermissions()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from . import ServiceHub

from redis import Redis
from redis.exceptions import RedisError


def handleRedisExceptions(service, error):
    if isinstance(error, RedisError):
        return ServiceHub.UNAVAILABLE


@ServiceHub.register("redis", handleRedisExceptions, maxfailures=5)
class RedisService(Redis):
    def __init__(self):
        from tools.config import Config
        conf = Config["sync"].get("connection", {})
        conf["decode_responses"] = True
        Redis.__init__(self, **conf)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

from argparse import ArgumentParser

from . import Cli

SUCCESS = 0
ERR_DECLINE = 1  # User declined prompt
ERR_USR_ABRT = 2  # User aborted
ERR_NO_LDAP = 3  # LDAP not available
ERR_GENERIC = 4  # Something went wrong
ERR_NO_USER = 5  # LDAP User not found
ERR_AMBIG = 6  # Request was ambiguous
ERR_DB = 7  # Error occurred when communicating with the database
ERR_CONFLICT = 8  # Target DB user is associated with another LDAP object
ERR_COMMIT = 9  # Error during database commit
ERR_INVALID_DATA = 10  # User data check failed
ERR_SETUP = 11  # Error during user setup

_ldapObjectColor = {
    "contact": "blue",
    "group": "cyan",
    "user": "green",
    }


def _getv(cli, prompt="", default="", secret=False):
    from getpass import getpass
    v = (getpass if secret else cli.input)("{}{}: ".format(prompt, " ["+str(default)+"]" if default is not None else ""))
    return default if v == "" else v


def _geti(cli, prompt="", default=0):
    res = None
    while res is None:
        res = _getv(cli, prompt, default)
        try:
            res = int(res)
        except:
            res = None
    return res


def _getc(cli, prompt="", default="", choices=(), getter=_getv):
    res = None
    while res is None:
        res = getter(cli, prompt, default)
        if res in choices:
            return res


def _getl(cli, prompt="", defaults=[]):
    cli.print(prompt+":")
    values = []
    defiter = (d for d in defaults)
    try:
        while True:
            val = _getv(cli, "", next(defiter, ""))
            if val == "":
                raise EOFError
            values.append(val)
    except EOFError:
        cli.print("[Done]")
    return values


def _getUsernameOrg(username):
    if "@" not in username:
        return None
    from orm.domains import Domains
    domain = Domains.query.filter(Domains.domainname == username.split("@", 1)[1]).with_entities(Domains.orgID).first()
    if domain is None:
        return None
    return domain.orgID


def _getOrgID(spec):
    from orm.domains import Orgs
    if spec == "0":
        return 0
    if Orgs.query.filter(Orgs.ID == spec).count() == 1:
        return int(spec)
    org = Orgs.query.filter(Orgs.name == spec).with_entities(Orgs.ID).first()
    if org is None:
        raise ValueError("Organization '{}' not found".format(spec))
    return org.ID


def _getOrgIDs(args):
    from orm.domains import OrgParam
    if "organization" not in args or not args.organization:
        return (0,)
    if "*" in args.organization:
        return [0]+OrgParam.ldapOrgs()
    return tuple(_getOrgID(spec) for spec in args.organization)


def _getOrgName(orgID):
    if orgID == 0:
        return "(no organization)"
    from orm.domains import Orgs
    org = Orgs.query.filter(Orgs.ID == orgID).with_entities(Orgs.name).first()
    if org is None:
        return "(unknown organozation)"
    return org.name


def _isSuccess(code):
    return code//100 == 2


def _userOrgFilter(args, orgIDs=None):
    from orm.users import Users
    return (Users.orgID.in_(orgIDs or _getOrgIDs(args)),)


def cliLdapInfo(args):
    cli = args._cli
    from services import Service
    for orgID in _getOrgIDs(args):
        with Service("ldap", orgID, errors=Service.SUPPRESS_INOP) as ldap:
            cli.print("Successfully connected to {}:{} as {}".format(cli.col(ldap.conn.server.host, attrs=["bold"]),
                                                                     cli.col(ldap.conn.server.port, attrs=["dark"]),
                                                                     ldap._config["connection"].get("bindUser", "<anonymous>")))


def _getCandidate(cli, expr, ldap):
    try:
        candidate = ldap.getUserInfo(ldap.unescapeFilterChars(expr))
        if candidate is not None:
            return candidate
    except Exception:
        pass
    matches = ldap.searchUsers(expr)
    if len(matches) == 0:
        cli.print(cli.col("Could not find LDAP object matching '{}'".format(expr), "red"))
        return None
    if len(matches) == 1:
        return matches[0]
    for match in matches:
        if match.email == expr:
            return match
    cli.print(cli.col(f"'{expr}' is ambiguous: ", "red"))
    cli.print(cli.col("\n  ".join(match.email for match in matches), "yellow"))
    return None


def _getCandidates(expr, ldap):
    candidate = ldap.getUserInfo(ldap.unescapeFilterChars(expr))
    return [candidate] if candidate is not None else ldap.searchUsers(expr)


def _downsync(args, user, externID=None, syncMembers=False, **kwargs):
    from tools.ldap import downsyncGroup, downsyncUser
    cli = args._cli

    cli.print("Synchronizing {}...".format(user.username), end="", flush=True)

    isGroup = user.properties["displaytypeex"] == 1
    if not isGroup:
        message, code = downsyncUser(user, externID)
    else:
        from orm.mlists import MLists
        mlist = MLists.query.filter(MLists.listname == user.username).first()
        if not mlist:
            cli.print(cli.col(f"Failed to synchronize {user.username}: no such group"))
            return False
        message, code = downsyncGroup(mlist, externID)
    cli.print(cli.col(message, "green" if _isSuccess(code) else "red"))
    if syncMembers and isGroup:
        _syncGroupMembers(args, mlist.user.orgID, mlist.user.externID)
    return _isSuccess(code)


def _importContact(args, candidate, ldap, orgID, syncExisting=False, **kwargs):
    from tools.ldap import importContact
    styleGood = {"color": "green"}
    styleNeutral = {"attrs": ["dark"]}
    styleBad = {"color": "red"}

    cli = args._cli
    cli.print("Importing {}...".format(candidate.email), end="", flush=True)
    synced, imported, errors = importContact(candidate, ldap, orgID, syncExisting)
    cli.print(cli.col(f"{len(synced)} synced", **styleGood if synced else styleNeutral)+", ", end="")
    cli.print(cli.col(f"{len(imported)} imported", **styleGood if imported else styleNeutral)+", ", end="")
    cli.print(cli.col(f"{len(errors)} failed", **styleBad if errors else styleNeutral))
    for error in errors:
        cli.print(cli.col("  "+error[0], "yellow"))
    return len(synced)+len(imported), len(errors)


def _importUser(args, candidate, ldap, **kwargs):
    from tools.ldap import importUser
    cli = args._cli

    cli.print("Importing {} {}...".format(candidate.type, candidate.email), end="", flush=True)
    result, code = importUser(candidate, ldap, args.force, args.lang)
    if not _isSuccess(code):
        cli.print(cli.col(result, "red"))
        return 0, 1
    cli.print(cli.col(f"imported as user #{result.ID}", "green"))
    return 1, 0


def _importGroup(args, candidate, ldap, syncMembers=False, **kwargs):
    from tools.ldap import importGroup
    cli = args._cli

    cli.print("Importing {}...".format(candidate.email), end="", flush=True)
    result, code = importGroup(candidate, ldap, args.force)
    if not _isSuccess(code):
        cli.print(cli.col(result, "red"))
        return 0, 1
    cli.print(cli.col(f"imported as user #{result.user.ID}/group #{result.ID}", "green"))
    if syncMembers:
        _syncGroupMembers(args, result.user.orgID, candidate.ID)
    return 1, 0


def _import(args, candidate, ldap, orgID, **kwargs):
    cli = args._cli
    if candidate.type == "contact":
        return _importContact(args, candidate, ldap, orgID, **kwargs)
    targetOrg = _getUsernameOrg(candidate.email)
    if targetOrg != orgID:
        cli.print(cli.col("Skipping '{}': invalid domain".format(candidate.email), "red"))
        return 0, 1
    if candidate.type == "group":
        return _importGroup(args, candidate, ldap, **kwargs)
    if candidate.type == "user":
        return _importUser(args, candidate, ldap, **kwargs)
    cli.print(cli.col(f"Unknown object type {candidate.type}"))


def _syncGroupMembers(args, orgID, groupID=None):
    from orm.users import Users
    from services import Service
    from tools.ldap import syncGroupMembers
    cli = args._cli

    users = {user.externID for user in
             Users.query.filter(Users.orgID == orgID, Users.externID != None).with_entities(Users.externID)}

    with Service("ldap", orgID) as ldap:
        ldapgroups = [ldap.getUserInfo(groupID)] if groupID else ldap.searchUsers(types=("group",))
        for ldapgroup in ldapgroups:
            cli.print(f"Synchronizing members of group {ldapgroup.email}...", end="", flush=True)
            add, remove = syncGroupMembers(orgID, ldapgroup, ldap, users)
            if None in (add, remove):
                cli.print(cli.col("group not found", attrs=["dark"]))
            else:
                cli.print(cli.col(f"{add} added, {remove} removed" if add+remove else "no changes", "green"))


def _downsyncOrg(args, orgID, ldap):
    cli = args._cli
    from orm.domains import Domains
    from orm.users import Users
    domainnames = [d.domainname for d in Domains.query.filter(Domains.orgID == orgID).with_entities(Domains.domainname)]
    if len(domainnames) == 0:
        cli.print(cli.col("Organization '{}' has no domains - skipping.".format(_getOrgName(orgID)), "yellow"))
        return (0, 0)
    synced = set()
    success = failed = 0
    for user in Users.query.filter(Users.orgID == orgID, Users.externID != None):
        result = _downsync(args, user)
        if user.status != Users.CONTACT:
            synced.add(user.externID)
        success += result
        failed += not result

    if args.complete:
        from services import Service
        with Service("ldap", orgID) as ldap:
            candidates = [candidate for candidate in ldap.searchUsers() if candidate.ID not in synced]
            for candidate in candidates:
                os, of = _import(args, candidate, ldap, orgID, syncExisting=False)
                success += os
                failed += of
    _syncGroupMembers(args, orgID)
    return (success, failed)


def _downsyncSpecific(args, orgIDs):
    if "user" not in args or not args.user:
        return (0, 0)
    cli = args._cli
    from orm.users import Users
    success = failed = 0
    for username in args.user:
        user = Users.query.filter(Users.username == username).first()
        if user:
            result = _downsync(args, user, syncMembers=True)
            success += result
            failed += not result
        else:
            from services import Service
            with Service("ldap", orgIDs[0]) as ldap:
                candidate = _getCandidate(cli, username, ldap)
                if not candidate:
                    failed += 1
                    continue
                os, of = _import(args, candidate, ldap, orgIDs[0], syncExisting=True, syncMembers=True)
                success += os
                failed += of
    return success, failed


def cliLdapDownsync(args):
    cli = args._cli
    cli.require("DB")
    from orm.users import Aliases, Users
    from services import Service
    Aliases.NTactive(False)
    Users.NTactive(False)
    orgIDs = _getOrgIDs(args)
    success, failed = _downsyncSpecific(args, orgIDs)
    if not success+failed:
        for orgID in orgIDs:
            with Service("ldap", orgID) as ldap:
                os, of = _downsyncOrg(args, orgID, ldap)
                success += os
                failed += of
    Aliases.NTactive(True)
    Users.NTactive(True)
    cli.print(cli.col("{} synchronized, {} failed".format(success, failed), attrs=["dark"]))


def cliLdapSearch(args):
    def typename(match):
        color = "yellow" if match.error else _ldapObjectColor.get(match.type, "magenta")
        return col(match.type, color)

    def col(text, *pargs, **kwargs):
        return cli.col(text, *pargs, **kwargs) if pretty else text

    cli = args._cli
    pretty = args.format == "pretty"
    from services import Service
    from .common import Table
    orgIDs = _getOrgIDs(args)
    types = args.types.split(",") if args.types else None
    for orgID in orgIDs:
        with Service("ldap", orgID) as ldap:
            if len(orgIDs) > 1:
                cli.print(cli.col(_getOrgName(orgID), "green"))
            matches = ldap.searchUsers(args.query, limit=args.max_results or None, pageSize=args.page_size,
                                       filterIncomplete=not args.all, types=types)
            hasErr = any(match.error is not None for match in matches)
            data = [(col(ldap.escape_filter_chars(match.ID), attrs=["bold"]), typename(match),
                     match.email if match.email else col("N/A", "red"), match.name,
                     col(match.error or "", "yellow"))
                    for match in matches if match.ID]
            table = Table(data, ("ID", "Type", "E-Mail", "Name", "Note" if hasErr or not pretty else ""),
                          empty=cli.col("(No results)", attrs=["dark"]) if pretty else None)
            table.dump(cli, args.format)
            if len(matches) and not pretty:
                cli.print(cli.col("({} result{})".format(len(matches), "s" if len(matches) != 1 else ""), attrs=["dark"]))


def cliLdapCheck(args):
    cli = args._cli
    cli.require("DB")
    from services import Service, ServiceUnavailableError
    from time import time
    from orm import DB
    from orm.users import Users
    users = Users.query.filter(Users.externID != None, *_userOrgFilter(args))\
                       .with_entities(Users.ID, Users.username, Users.externID, Users.maildir, Users.orgID).all()
    if len(users) == 0:
        cli.print("No imported users found. You can import users using 'ldap downsync <name>' or 'ldap downsync --complete'.")
        return
    cli.print("Checking {} user{}...".format(len(users), "" if len(users) == 1 else "s"))
    count, last = 0, time()
    orphaned = []
    for user in users:
        try:
            with Service("ldap", user.orgID) as ldap:
                if ldap.getUserInfo(user.externID) is None:
                    orphaned.append(user)
                count += 1
                if time()-last > 1:
                    last = time()
                    cli.print("\t{}/{} checked ({:.0f}%), {} orphaned"
                              .format(count, len(users), count/len(users)*100, len(orphaned)))
        except ServiceUnavailableError:
            cli.print(cli.col("\tFailed to check user '"+user.username+"' - LDAP not available", "red"))
    if len(orphaned) == 0:
        cli.print("Everything is ok")
        return
    cli.print("LDAP entries of the following users could not be found:")
    for user in orphaned:
        cli.print("\t"+user.username)
    if args.remove:
        if args.yes or cli.confirm("Delete all orphaned users? [y/N]: ") == Cli.SUCCESS:
            cli.print("Unloading exmdb stores...")
            if len(orphaned):
                homeserver = None
                users = Users.query.filter(Users.ID.in_(orphan.ID for orphan in orphaned)).order_by(Users.homeserverID).all()
                index = 0
                while index < len(users):
                    try:
                        with Service("exmdb") as exmdb:
                            if homeserver != users[index].homeserverID:  # Reuse the exmdb client for users on the same server
                                user = users[index]
                                client = exmdb.ExmdbQueries(exmdb.host if user.homeserverID == 0 else user.homeserver.hostname,
                                                            exmdb.port, user.maildir, True)
                                homeserver = user.homeserverID
                            while index < len(users) and users[index].homeserverID == homeserver:
                                client.unloadStore(users[index].maildir)
                                index += 1
                    except ServiceUnavailableError:
                        cli.print(cli.col("Failed to unload store: exmdb service not available", "yellow"))
                        index += 1
                if args.remove_maildirs:
                    import shutil
                    cli.print("Removing mail directories...")
                    for user in orphaned:
                        shutil.rmtree(user.maildir, ignore_errors=True)
                for user in users:
                    user.delete()
            DB.session.commit()
            cli.print("Deleted {} user{}".format(len(users), "" if len(users) == 1 else "s"))
            return
    return ERR_NO_USER


def cliLdapDump(args):
    cli = args._cli
    ldapArgs = (args.organization,) if args.organization else ()
    from services import Service
    results = 0
    with Service("ldap", *ldapArgs) as ldap:
        for expr in args.user:
            for candidate in _getCandidates(expr, ldap):
                cli.print(cli.col("ID: "+ldap.escape_filter_chars(candidate.ID), attrs=["bold"]))
                cli.print(str(ldap.dumpUser(candidate.ID)))
                results += 1
    cli.print(cli.col("({} result{})".format(results, "s" if results != 1 else ""), attrs=["dark"]))


def _applyTemplate(index, conf):
    conf["users"] = conf.get("users", {})
    if index == 1:  # AD
        conf["objectID"] = "objectGUID"
        conf["users"]["aliases"] = "proxyAddresses"
        conf["users"]["displayName"] = "displayName"
        conf["users"]["username"] = "mail"
    elif index == 2:  # OpenLDAP
        conf["objectID"] = "entryUUID"
        conf["users"]["aliases"] = "mailAlternativeAddress"
        conf["users"]["displayName"] = "displayname"
        conf["users"]["username"] = "mailPrimaryAddress"


def _checkConn(cli, connfig):
    cli.print(cli.col("Checking connectivity...", attrs=["dark"]), end="", flush=True)
    from services.ldap import LdapService
    try:
        LdapService.testConnection({"connection": connfig}, active=False)
    except Exception as exc:
        cli.print(cli.col("\nConnection check failed: "+" - ".join(str(arg) for arg in exc.args), "red"))
        res = cli.choice("(a)bort, (c)ontinue anyway, (e)dit configuration? [e]: ", "ace", "e")
        if res in (None, "a"):
            raise KeyboardInterrupt
        return res == "c"
    cli.print(cli.col("success!", "green", attrs=["dark"]))
    return True


def _getConf(cli, old):
    conf = {"connection": {}, "users": {"filters": [], "searchAttributes": []}}
    connected = False
    connfig = old.get("connection", {}).copy()
    while not connected:
        oldpw = "[***]" if connfig.get("bindPass") else "[]"
        connfig["server"] = _getv(cli, "URL of the LDAP server(s)", connfig.get("server", ""))
        connfig["bindUser"] = _getv(cli, "Username for access", connfig.get("bindUser"), )
        connfig["bindPass"] = _getv(cli, "Password for access "+oldpw, None, True) or connfig.get("bindPass")
        connfig["starttls"] = _getc(cli, "Use StartTLS connection",
                                    "y" if connfig.get("starttls") else "n", ("y", "n")) == "y"
        connected = _checkConn(cli, connfig)
    conf["connection"] = connfig
    conf["baseDn"] = _getv(cli, "Search base for user lookup/searches", old.get("baseDn", ""))
    users = old.get("users", {})
    oldtempl = users.get("templates", ())
    res = _getc(cli, "Choose a template:\n 0: No template\n 1: ActiveDirectory\n 2: OpenLDAP\n",
                1 if "ActiveDirectory" in oldtempl else 2 if "OpenLDAP" in oldtempl else 0, range(2), _geti)
    conf["users"]["templates"] = [] if res == 0 else ["common", "ActiveDirectory" if res == 1 else "OpenLDAP"]
    if res != 0 and cli.confirm("Apply default template parameters? [y/N]: ") == Cli.SUCCESS:
        _applyTemplate(res, old)
        users = old.get("users", {})
    conf["objectID"] = _getv(cli, "Attribute containing unique object ID", old.get("objectID"))
    conf["users"]["username"] = _getv(cli, "Attribute containing e-mail address of a user", users.get("username", ""))
    conf["users"]["displayName"] = _getv(cli, "Attribute containing name of a user", users.get("displayName", ""))
    conf["users"]["aliases"] = _getv(cli, "Attribute containing alternative e-mail addresses", users.get("aliases", ""))
    conf["users"]["defaultQuota"] = _geti(cli, "Default storage quota for imported users (0=unlimited)",
                                          users.get("defaultQuota", 0))
    conf["users"]["filter"] = _getv(cli, "Enter filter expression for user search", users.get("filter", ""))
    conf["users"]["contactFilter"] = _getv(cli, "Enter filter expression for contact search", users.get("contactFilter", ""))
    conf["users"]["searchAttributes"] = _getl(cli, "Enter attributes used for searching (one per line)",
                                              users.get("searchAttributes", []))
    if not conf["users"]["defaultQuota"]:
        conf["users"].pop("defaultQuota")
    return conf


def _cliLdapGetConf(args):
    if args.organization:
        from orm.domains import OrgParam
        args.organization = _getOrgID(args.organization)
        old = OrgParam.loadLdap(args.organization)
    if not args.organization or not old:
        from tools import mconf
        old = mconf.LDAP
    return old


def _cliLdapSaveConf(args, conf):
    if not args.organization:
        from tools import mconf
        return mconf.dumpLdap(conf)
    from orm.domains import OrgParam
    OrgParam.saveLdap(args.organization, conf)


def _cliLdapConfigure(args):
    from services import ServiceHub
    try:
        cli = args._cli
        if args.delete:
            if not args.organization:
                cli.print(cli.col("Cannot delete default configuration"))
                return 2
            orgID = _getOrgID(args)
            from orm.domains import OrgParam
            OrgParam.wipeLdap(orgID)
            cli.print("Configuration deleted.")
            ServiceHub.load("ldap", orgID, force_reload=True)
            return

        from services.ldap import LdapService
        LdapService.init()
        old = _cliLdapGetConf(args)
        while True:
            new = _getConf(cli, old)
            cli.print("Checking new configuration...")
            error = LdapService.testConfig(new)
            if error is None:
                cli.print("Configuration successful.")
                _cliLdapSaveConf(args, new)
                cli.print("Configuration saved" if error is None else ("Failed to save configuration: "+error))
                break
            cli.print(cli.col(error, "yellow"))
            action = _getc(cli, "Restart configuration? (r=Restart, a=Amend, s=Save anyway, q=quit)",
                           "a", ("y", "a", "s", "q"))
            if action == "s":
                _cliLdapSaveConf(args, new)
            if action in "sq":
                break
            if action == "a":
                old = new
    except (KeyboardInterrupt, EOFError):
        cli.print(cli.col("\nAborted."))
        return 1
    except ValueError as err:
        cli.print(cli.col(err.args[0], "red"))
    ldapArgs = (args.organization,) if args.organization else ()
    ServiceHub.load("ldap", *ldapArgs, force_reload=True)


def cliLdapReload(args):
    from services import ServiceHub
    cli = args._cli
    ldapArgs = (args.organization,) if args.organization else ()
    res = ServiceHub.load("ldap", *ldapArgs, force_reload=True)
    cli.print("Reload successful" if res.state == ServiceHub.LOADED else cli.col("Reload failed", "red"))
    return int(res.state != ServiceHub.LOADED)


def _cliOrgspecCompleter(prefix, **kwargs):
    from orm.domains import Orgs
    return (org.name for org in Orgs.query.filter(Orgs.name.ilike(prefix+"%")).with_entities(Orgs.name).all())


def _cliLdapParserSetup(subp: ArgumentParser):
    sub = subp.add_subparsers()
    check = sub.add_parser("check", help="Check LDAP objects of imported users still exist")
    check.set_defaults(_handle=cliLdapCheck)
    check.add_argument("-m", "--remove-maildirs", action="store_true", help="When deleting users, also remove their mail "
                                                                            "directories from disk")
    check.add_argument("-o", "--organization", metavar="ORGSPEC", action="append",
                       help="Use organization specific LDAP connection").completer = _cliOrgspecCompleter
    check.add_argument("-r", "--remove", action="store_true", help="Prompt for user deletion if orphaned users exist")
    check.add_argument("-y", "--yes", action="store_true", help="Do not prompt for user deletion (only with -r)")
    configure = sub.add_parser("configure", help="Run interactive LDAP configuration")
    configure.set_defaults(_handle=_cliLdapConfigure)
    configure.add_argument("-d", "--delete", action="store_true", help="Do not configure anything, but delete configuration")
    configure.add_argument("-o", "--organization", metavar="ORGSPEC", help="Use organization specific LDAP connection")\
        .completer = _cliOrgspecCompleter
    downsync = sub.add_parser("downsync", help="Import or update users from ldap")
    downsync.set_defaults(_handle=cliLdapDownsync)
    downsync.add_argument("user", nargs="*", help="LDAP ID or user search query string. If omitted, all users linked to an "
                                                  "LDAP object are updated.")
    downsync.add_argument("-c", "--complete", action="store_true", help="Import/update all users in the ldap tree")
    downsync.add_argument("-f", "--force", action="store_true", help="Force synchronization of unassociated users")
    downsync.add_argument("-l", "--lang", help="Default language for imported users")
    downsync.add_argument("-o", "--organization", metavar="ORGSPEC", action="append",
                          help="Use organization specific LDAP connection").completer = _cliOrgspecCompleter
    downsync.add_argument("-p", "--page-size", type=int, default=1000, help="Page size when downloading users")
    dump = sub.add_parser("dump", help="Dump LDAP object")
    dump.set_defaults(_handle=cliLdapDump)
    dump.add_argument("-o", "--organization", metavar="ORGSPEC", help="Use organization specific LDAP connection")\
        .completer = _cliOrgspecCompleter
    dump.add_argument("user", nargs="+", help="User ID or search query string")
    info = sub.add_parser("info", help="Check LDAP status")
    info.set_defaults(_handle=cliLdapInfo)
    info.add_argument("-o", "--organization", metavar="ORGSPEC", action="append",
                      help="Use organization specific LDAP connection").completer = _cliOrgspecCompleter
    reload = sub.add_parser("reload", help="Reload LDAP configuration")
    reload.set_defaults(_handle=cliLdapReload)
    reload.add_argument("-o", "--organization", metavar="ORGSPEC", help="Use organization specific LDAP connection")\
        .completer = _cliOrgspecCompleter
    search = sub.add_parser("search", help="Search LDAP tree")
    search.set_defaults(_handle=cliLdapSearch)
    search.add_argument("query", nargs="?", help="Optional search query, omit to return all users")
    search.add_argument("-a", "--all", action="store_true", help="Also show users that cannot be imported")
    search.add_argument("--format", choices=("csv", "json-flat", "json-structured", "pretty"), help="Set output format",
                        metavar="FORMAT", default="pretty")
    search.add_argument("-n", "--max-results", type=int, default=0,
                        help="Maximum number of results or 0 to disable limit (default: 0)")
    search.add_argument("-o", "--organization", metavar="ORGSPEC", action="append",
                        help="Use organization specific LDAP connection").completer = _cliOrgspecCompleter
    search.add_argument("-p", "--page-size", type=int, default=1000, help="Page size when downloading users")
    search.add_argument("-t", "--types", help="Comma separated list of results types to search for")


@Cli.command("ldap", _cliLdapParserSetup, help="LDAP configuration, diagnostics and synchronization")
def cliLdap(args):
    return cliLdapInfo(args)



#!/usr/bin/env python3.10
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2020 grommunio GmbH
"""
Main file to execute.

Can be used to run the API in stand-alone mode by directly executing the file or
in combination with a WSGI server using the API object as callable
"""

if __name__ == '__main__':
    import os
    import sys
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    from cli import Cli
    res = Cli().execute()
    sys.exit(res)
else:
    from api.core import API  # Export to uwsgi server
    from cli import Cli
    from endpoints import *  # Register all endpoints
    from tools import config
    Cli.initLogging()
    error = config.validate()
    if error:
        raise TypeError("Invalid configuration found - aborting ({})".format(error))
    if not config.Config["tasq"].get("disabled", False):
        import uwsgi
        import uwsgidecorators
        from tools.tasq import TasQServer

        @uwsgidecorators.postfork
        def enableTasQ():
            TasQServer.start()
            uwsgi.atexit = TasQServer.stop



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH


def domainFilter(domainSpec, *filters):
    from orm.domains import Domains
    from sqlalchemy import and_
    return and_(True if domainSpec is None else
                Domains.ID == domainSpec if domainSpec.isdigit() else
                Domains.domainname.ilike(domainSpec+"%"), *filters)


def domainCandidates(domainSpec, *filters):
    from orm.domains import Domains
    return Domains.query.filter(domainFilter(domainSpec, *filters))


def userFilter(userSpec, *filters):
    from orm.users import Users
    from sqlalchemy import and_
    return and_(True if userSpec is None else
                Users.ID == userSpec if userSpec.isdigit() else
                Users.username.ilike(userSpec+"%"), *filters)


def userCandidates(userSpec, *filters):
    from orm.users import Users
    return Users.query.filter(userFilter(userSpec, *filters))


def userspecAutocomp(prefix, **kwargs):
    from . import Cli
    if Cli.rlAvail:
        from orm.users import Users
        return (user.username for user in userCandidates(prefix).with_entities(Users.username))
    else:
        return ()


class NotFound(dict):
    pass


def getKey(c, keyspec):
    if keyspec:
        for key in keyspec:
            c = c.get(key, NotFound()) if key else c
    return c


def proptagCompleter(prefix, addSuffix="", **kwargs):
    from tools.constants import PropTags
    PropTags.lookup(None)
    c = []
    if prefix == "" or prefix[0].islower():
        c += [tag.lower()+addSuffix for value, tag in PropTags._lookup.items() if isinstance(value, int)]
    if prefix == "" or prefix[0].isupper():
        c += [tag.upper()+addSuffix for value, tag in PropTags._lookup.items() if isinstance(value, int)]
    if prefix == "" or prefix[0] == "0" and (len(prefix) <= 2 or not prefix[2:].isupper()):
        c += ["0x{:08x}{}".format(value, addSuffix) for value in PropTags._lookup.keys() if isinstance(value, int)]
    if prefix == "" or prefix[0] == "0" and (len(prefix) <= 2 or not prefix[2:].islower()):
        c += ["0x{:08X}{}".format(value, addSuffix) for value in PropTags._lookup.keys() if isinstance(value, int)]
    if prefix == "" or prefix.isnumeric():
        c += [str(value)+addSuffix for value in PropTags._lookup.keys() if isinstance(value, int)]
    return c


class Table:
    """Helper class for pretty printing of tables."""
    class Styled:
        """Class to manage style information of a table cell"""
        stylemarker = None

        def __init__(self, data, align='a', color=None, on_color=None, attrs=[]):
            """Associate style information with data.

            Parameters
            ----------
            data : any
                Data to display
            align : str, optional
                Alignment, can be one of 'a' (auto), 'c' (center), 'l' (left) or 'r' (right).
                Automatic alignment chooses 'r' for numbers and 'l' for everything else.
                The default is 'a'.
            color : str, optional
                Color to apply to the text. The default is None.
            on_color : str, optional
                Background color of the text. The default is None.
            attrs : [str], optional
                List of additional style attributes. The default is [].
            """
            self._init()
            self.align = align if align != "a" else "r" if type(data) in (int, float) else "l"
            self.raw = data
            self.data = str(data).expandtabs()
            self.color = color
            self.on_color = None
            self.attrs = attrs
            self.width = self._width()

        @classmethod
        def _init(cls):
            """Initialize stylemarker re."""
            if cls.stylemarker is not None:
                return
            import re
            cls.stylemarker = re.compile("\x1b\\[[\\d]{1,2}m")

        def _width(self):
            """Return effective width of the string (without style markers and expanded tabs)."""
            return len(self.stylemarker.sub("", self.data).expandtabs())

        def print(self, cli, width, last):
            """Print styled data into string.

            Parameters
            ----------
            cli : Cli
                Cli providing style formatting.
            width : int
                Width of the cell to fill
            last : bool
                Whether this is the last cell of the row

            Returns
            -------
            data : str
                Cell content
            """
            pad = width-self.width
            data = cli.col(self.data, self.color, self.on_color, self.attrs)
            if self.align == "r":
                data = " "*pad+data
            elif self.align == "c":
                data = " "*(pad//2)+data+(" "*((pad+1)//2) if not last else "")
            elif not last:
                data += " "*pad
            return data

    def __init__(self, data, header=None, colsep="  ", empty=None):
        """Create table from data.

        Parameters
        ----------
        data : [[any]]
            Matrix of contents
        header : [any], optional
            Table header. The default is None.
        colsep : str, optional
            Column separator. The default is "  ".
        empty : str, optional
            Text to display when table does not contain data. The default is None.
        """
        self.data = [[self._styled(cell) for cell in row] for row in data] if data else None
        self.header = [self._styled(col, "l", attrs=["underline"]) for col in header] if header else None
        self.empty = empty
        if not (header or data):
            return
        self.colsep = colsep
        head = self.header or self.data[0]
        self.columns = max(max(len(row) for row in data) if data else 0, len(head))
        self.colwidth = tuple(head[i].width if i < len(head) else 0 for i in range(self.columns))
        if self.data:
            for line in self.data:
                self.colwidth = tuple(max(self.colwidth[i], line[i].width if i < len(line) else 0)
                                      for i in range(self.columns))

    @classmethod
    def _styled(cls, data, *args, **kwargs):
        """Augment data with style information.

        Parameters
        ----------
        data : any
            Data to wrap
        *args : any
            Arguments passed on to Styled constructor
        **kwargs : any
            Keyword arguments passed on to Styled constructor.

        Returns
        -------
        Styled
            Data with style information
        """
        return data if isinstance(data, cls.Styled) else cls.Styled(data, *args, **kwargs)

    def printline(self, cli, line):
        """Print a single row of data.

        Parameters
        ----------
        cli : Cli
            Cli providing printing functionality
        line : [Styled]
            List of cells to print
        """
        cli.print(self.colsep.join(line[i].print(cli, self.colwidth[i], i == self.columns-1) for i in range(len(line))))

    def print(self, cli):
        """Print the table.

        Parameters
        ----------
        cli : Cli
            Cli providing printing functionality.
        """
        if not self.data and self.empty:
            cli.print(self.empty)
            return
        if self.header:
            self.printline(cli, self.header)
        if self.data:
            for line in self.data:
                self.printline(cli, line)

    def csv(self, cli):
        """Output table as csv.

        Parameters
        ----------
        cli : Cli
            Cli providing printing functionality.
        """
        import csv
        if not (self.header or self.data):
            return
        header = [cell.raw for cell in self.header] if self.header else [str(i) for i in range(len(self.data[0]))]
        writer = csv.DictWriter(cli.stdout, fieldnames=header, delimiter=self.colsep[0] or ",")
        if self.header:
            writer.writeheader()
        if not self.data:
            return
        for row in self.data:
            writer.writerow({name: value.raw for name, value in zip(header, row)})

    def json(self, cli, structured):
        """Output table as JSON

        Parameters
        ----------
        cli : Cli
            Cli providing printing functionality.
        structured : bool
            Whether to output data as structured JSON or array-of-arrays
        """
        import json
        if not self.data:
            cli.print("[]")
            return
        header = [cell.raw for cell in self.header] if self.header else [str(i) for i in range(len(self.data[0]))]
        data = [{name: value.raw for name, value in zip(header, row)} for row in self.data] if structured else\
               [[cell.raw for cell in row] for row in self.data]
        cli.print(json.dumps(data, default=lambda x: str(x), separators=(",", ":")))

    def dump(self, cli, format):
        """Dump table contents in specified format

        Parameters
        ----------
        cli : Cli
            Cli providing printing functionality.
        format : str
            Output format. Can be one of `csv`, `json-flat` and `json-structured`. Everything else will print the table.
        """
        if format == "csv":
            self.csv(cli)
        elif format in ("json-flat", "json-structured"):
            self.json(cli, format == "json-structured")
        else:
            self.print(cli)



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2022 grommunio GmbH

from . import Cli, InvalidUseError, ArgumentParser


def _dumpServer(cli, server):
    cli.print(cli.col("{} ({})".format(server.hostname, server.ID), attrs=["bold"]))
    cli.print("  ID: "+str(server.ID))
    cli.print("  hostname: "+server.hostname)
    cli.print("  extname: "+server.extname)
    cli.print("  domains: "+str(server.domains))
    cli.print("  users: "+str(server.users))


def _sanitizeData(data):
    cliargs = {"_handle", "_cli", "serverspec"}
    return {key: value for key, value in data.items() if value is not None and key not in cliargs}


def _serverCandidates(serverspec):
    from orm.misc import Servers
    if serverspec is None:
        return Servers.query
    filter = Servers.ID == serverspec if serverspec.isdigit() else Servers.hostname == serverspec
    return Servers.query.filter(filter)


def cliServersCreate(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import DB, Servers
    data = _sanitizeData(args.__dict__)
    error = Servers.checkCreateParams(data)
    if error:
        cli.print(cli.col("Could not create server: "+error, "red"))
        return 1
    server = Servers(data)
    DB.session.add(server)
    DB.session.commit()
    _dumpServer(cli, server)


def cliServersDelete(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    servers = _serverCandidates(args.serverspec).all()
    if len(servers) == 0:
        cli.print(cli.col("Server not found", "yellow"))
        return 1
    if len(servers) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.serverspec), "yellow"))
        return 2
    server = servers[0]
    if server.users+server.domains:
        cli.print(cli.col("Cannot delete server with users or domains", "red"))
        return 3
    DB.session.delete(servers[0])
    DB.session.commit()
    cli.print("Server deleted.")


def cliServersList(args):
    cli = args._cli
    cli.require("DB")
    from orm.misc import Servers
    query = _serverCandidates(args.serverspec)
    if "filter" in args and args.filter is not None:
        query = Servers.autofilter(query, {f.split("=", 1)[0]: f.split("=", 1)[1] for f in args.filter if "=" in f})
    if "sort" in args and args.sort is not None:
        query = Servers.autosort(query, args.sort)
    servers = query.all()
    for server in servers:
        cli.print("{}: {} ({})".format(server.ID, cli.col(server.hostname, attrs=["bold"]), server.extname))
    cli.print(cli.col("({} result{})".format(len(servers), "" if len(servers) == 1 else "s"), attrs=["dark"]))


def cliServersModify(args):
    cli = args._cli
    cli.require("DB")
    from orm import DB
    servers = _serverCandidates(args.serverspec).all()
    if len(servers) == 0:
        cli.print(cli.col("Server not found", "yellow"))
        return 1
    if len(servers) > 1:
        cli.print(cli.col("'{}' is ambiguous".format(args.serverspec), "yellow"))
        return 2
    server = servers[0]
    data = _sanitizeData(args.__dict__)
    server.fromdict(data)
    DB.session.commit()
    _dumpServer(cli, server)


def cliServersShow(args):
    cli = args._cli
    cli.require("DB")
    servers = _serverCandidates(args.serverspec).all()
    if len(servers) == 0:
        cli.print(cli.col("No servers found.", "yellow"))
        return 1
    for server in servers:
        _dumpServer(cli, server)


def _cliServerServerspecAutocomp(prefix, **kwargs):
    from orm.misc import Servers
    return (server.hostname for server in Servers.query.filter(Servers.hostname.ilike(prefix+"%")).all())


def _setupCliServersParser(subp: ArgumentParser):
    def addProperties(parser, init):
        parser.add_argument("-e", "--extname", required=init, help="External host name")
        parser.add_argument("-H", "--hostname", required=init, help="Internal host name")
    Cli.parser_stub(subp)
    sub = subp.add_subparsers()
    create = sub.add_parser("create", help="Create new server entry")
    create.set_defaults(_handle=cliServersCreate)
    addProperties(create, True)
    delete = sub.add_parser("delete", help="Remove server")
    delete.set_defaults(_handle=cliServersDelete)
    delete.add_argument("serverspec", help="Server ID or host name").completer = _cliServerServerspecAutocomp
    list = sub.add_parser("list", help="List servers")
    list.set_defaults(_handle=cliServersList)
    list.add_argument("serverspec", nargs="?", help="Server ID or prefix to match host name against")
    list.add_argument("-f", "--filter", nargs="*", help="Filter by attribute, e.g. -f ID=42")
    list.add_argument("-s", "--sort", nargs="*", help="Sort by attribute, e.g. -s hostname,desc")
    modify = sub.add_parser("modify", help="Modify server entry")
    modify.set_defaults(_handle=cliServersModify)
    modify.add_argument("serverspec", help="server ID or host name").completer = _cliServerServerspecAutocomp
    addProperties(modify, False)
    show = sub.add_parser("show", help="Show detailed information about a server")
    show.set_defaults(_handle=cliServersShow)
    show.add_argument("serverspec", help="Server ID or host name").completer = _cliServerServerspecAutocomp


@Cli.command("server", _setupCliServersParser, help="Multi-server management")
def cliServersStub(args):
    raise InvalidUseError()



# -*- coding: utf-8 -*-
# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2021 grommunio GmbH

#XXXfrom systemd.journal import Reader

class LogReader:
    """Central log reader class."""
    rreg = {}

    @classmethod
    def register(cls, name):
        """Class decorator to register a log reader at the factory.

        Parameters
        ----------
        name : str
            Name of the log reader

        Returns
        -------
        function
            Decorator function
        """
        def inner(obj):
            cls.rreg[name] = obj
            return obj
        return inner

    @classmethod
    def tail(cls, source, target, *args, **kwargs):
        """Get log tail.

        Automatically uses the correct log reader according to `source`.

        Parameters
        ----------
        source : str
            Name of the log source
        target : str
            Name of the log file or unit
        *args : any
            Arguments forwarded to the log reader
        **kwargs : any
            Keyword arguments forwarded to the log reader

        Raises
        ------
        ValueError
            `source` is not a registered log reader

        Returns
        -------
        list
            List of log file entries
        """
        if source not in cls.rreg:
            raise ValueError("Unknown source '{}'".format(source))
        return cls.rreg[source](target).tail(*args, **kwargs)


@LogReader.register("journald")
class JournaldReader:
    """Reader class four journald logs."""

    def __init__(self, unit):
        """Create journald reader

        Parameters
        ----------
        unit : str
            Name of the unit.
        """
        self.reader = Reader()
        self.reader.add_match(_SYSTEMD_UNIT=unit)

    @staticmethod
    def _entry(data):
        return dict(level=data["PRIORITY"],
                    message=data["MESSAGE"],
                    time=data["__REALTIME_TIMESTAMP"].strftime("%Y-%m-%d %H:%M:%S.%f"),
                    runtime=data["__MONOTONIC_TIMESTAMP"].timestamp.total_seconds())

    def tail(self, n=10, skip=0, after=None):
        """Get log tail.

        Parameters
        ----------
        n : int, optional
            Number of lines to return. The default is 10.
        skip : int, optional
            Number of lines to skip. The default is 0.
        after : datetime, optional
            Return all lines after given time point. Overrides `n` and `skip`. The default is None.

        Returns
        -------
        list
            List of log file entries
        """
        self.reader.seek_tail()
        if after is None:
            if skip > 0:
                self.reader.get_previous(skip)
            entries = reversed([self.reader.get_previous() for _ in range(n)])
            return [self._entry(entry) for entry in entries if len(entry) != 0]
        entries = []
        while True:
            entry = self.reader.get_previous()
            if len(entry) == 0 or entry["__REALTIME_TIMESTAMP"] <= after:
                break
            entries.append(self._entry(entry))
        return list(reversed(entries))
